------------------------------------------------------------------------
----                                                                ----
---- ATARI Falcon COMBEL compatible IP Core                         ----
----                                                                ----
---- This file is part of the SUSKA ATARI clone project.            ----
---- http://www.experiment-s.de                                     ----
----                                                                ----
---- Description:                                                   ----
---- Atari's COMBEL with all features to reach                      ----
---- ATARI Falcon compatibility.                                    ----
----                                                                ----
---- Top level file for use in systems on programmable chips.       ----
----                                                                ----
---- To Do:                                                         ----
---- -                                                              ----
----                                                                ----
---- Author(s):                                                     ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de     ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Copyright © 2009-2015 Wolfgang Foerster Inventronik GmbH.      ----
---- All rights reserved. No portion of this sourcecode may be      ----
---- reproduced or transmitted in any form by any means, whether    ----
---- by electronic, mechanical, photocopying, recording or          ----
---- otherwise, without my written permission.                      ----
----                                                                ----
------------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K9B  2009/12/24 WF
--   Initial Release.
-- Revision 2K15B  2015/12/24 WF
--   Several changes and fixes:
--   Added CSn_18B20 for the onewire device.
--   Added CSn_MP34DB01 for the microphone.
-- 

library work;
use work.COMBEL_PKG.all;
library ieee;
use ieee.std_logic_1164.all;

entity COMBEL_TOP is
    port (
        -- System and core control:
        RESET               : in std_logic;
        CPU_CLKSEL_16_8n    : out std_logic;
        BLT_CLKSEL_16_8n    : out std_logic;
        CLK_32              : in std_logic; -- Originally 32MHz.
        CLK_BLT             : in std_logic;
        CLK_8               : out std_logic;
        CLK_4               : out std_logic;
        KHz_500             : out std_logic; -- Clock for the MIDI ACIA and the keyboard ACIA (receiver).
        KHz_500W            : out std_logic; -- Clock for the transmitter of the keyboard ACIA.

        -- Adress and data bus:
        ADR_IN              : in std_logic_vector(31 downto 1);
        ADR_16MB_UP         : in std_logic; -- Select the fully 32 bit address space.
        ADR_OUT             : out std_logic_vector(31 downto 1);
        ADR_EN              : out std_logic; -- Used for the Blitter.
        DATA_IN             : in std_logic_vector(15 downto 0);
        DATA_OUT            : out std_logic_vector(15 downto 0);
        DATA_EN             : out std_logic;

        -- The RAM interface:
        RAS1n               : out std_logic;
        CAS1Hn              : out std_logic;
        CAS1Ln              : out std_logic;
        RAS0n               : out std_logic;
        CAS0Hn              : out std_logic;
        CAS0Ln              : out std_logic;
        WEn                 : out std_logic;
        RAM_ADR             : out std_logic_vector(26 downto 1);

        -- Bus control:
        RWn                 : in std_logic;
        ASn                 : in std_logic;
        UDSn                : in std_logic;
        LDSn                : in std_logic;

        DTACK_INn           : in std_logic;
        DTACK_OUTn          : out std_logic;

        -- 6800 peripheral control:
        VPAn                : out std_logic;
        VMAn                : in std_logic;

        -- Bus status:
        BERRn               : out std_logic;

        -- Processor function codes:
        FC                  : in std_logic_vector(2 downto 0);

        -- Bus arbitration control:
        BRn                 : out std_logic;
        BGIn                : in std_logic;
        BGOn                : out std_logic;
        BGAn                : out std_logic;

        -- Adress decoder stuff:
        -- In original COMBEL ther are only
        -- Pins for ROM2n, ROM3n and ROM4n.
        ROMSEL_FC_E0n       : in std_logic; -- Operating system's address space.
        ROM_6n              : out std_logic;  -- STE.
        ROM_5n              : out std_logic;  -- STE.
        ROM_4n              : out std_logic;  -- ST.
        ROM_3n              : out std_logic;  -- ST.
        ROM_2n              : out std_logic;
        ROM_1n              : out std_logic;
        ROM_0n              : out std_logic;

        N6850               : out std_logic;
        MFPCSn              : out std_logic;

        SNDCS               : out std_logic; -- STE: Sound chip select.
        SNDIR               : out std_logic; -- STE: Data flow direction control.
        FPUCS               : out std_logic; -- Floating point unit
        R8006               : out std_logic; -- Falcon's configuration register.

        -- Keyboard stuff:
        TOK                 : in std_logic; -- 'Transmit Ok', to KROK of the Keyboard connector.
        TID                 : in std_logic; -- To TXD of the keyboard ACIA, not used yet.

        -- VIDEL control signals:
        VREQ                : in std_logic; -- Video data request.
        EVENn_ODD           : in std_logic; -- Indicates the interlaced video frame.
        VCS                 : out std_logic; -- VIDEL chip select.
        VLD                 : out std_logic; -- VIDEL's data latch control.
        RDAT                : out std_logic; -- VIDEL's data latch control.
        WDAT                : out std_logic; -- VIDEL's data latch control.
        RAMH                : out std_logic; -- Indicates hi RAM access to the VIDEL.

        -- DMA control signal:
        BMODE               : in std_logic; -- '0' = 68030 bus timing '1' = 68000 bus timing.

        -- Real time clock:
        RTCCS               : out std_logic; -- Real time clock chip select.
        RTCAS               : out std_logic; -- Address strobe.
        RTCDS               : out std_logic; -- Data strobe.
        RTC_ACK             : in std_logic; -- Set to '1' if not used.

        -- Interrupt system:
        HINTn               : in std_logic; -- Horizontal interrupt.
        VINTn               : in std_logic; -- Vertical Interrupt.
        MFPINTn             : in std_logic;
        EINT1n              : in std_logic;
        EINT3n              : in std_logic;
        EINT5n              : in std_logic;
        EINT7n              : in std_logic;
        BINTn               : out std_logic; -- Blitter.
        AVECn               : out std_logic; -- Add-On over COMBEL.
        IACKn               : out std_logic; -- ST.
        IPLn                : out std_logic_vector(2 downto 0); -- STE only.

        -- IDE interface:
        IDE_RS0n            : out std_logic;
        IDE_RS1n            : out std_logic;
        IDE_IORDn           : out std_logic;
        IDE_IOWRn           : out std_logic;
        IDE_D_EN_INn        : out std_logic; -- In-Buffer control, Add-On over COMBEL.
        IDE_D_EN_OUTn       : out std_logic; -- Out-Buffer control, Add-On over COMBEL.

        -- SCC chip:
        SCCAB               : out std_logic;
        SCCRD               : out std_logic;
        SCCWR               : out std_logic;
        SCCIACKn            : out std_logic;
        SCCWAITn            : in std_logic;

        -- Joyport:
        JOY_RHn             : out std_logic;      -- Read only FF9202 high byte.
        JOY_RLn             : out std_logic;      -- Read only FF9202 low byte.
        JOY_WL              : out std_logic;      -- Write only FF9202 low byte.
        JOY_WEn             : out std_logic;      -- Write only FF9202 output enable.
        BUTTONn             : out std_logic;      -- Read only FF9000 low byte.       
        PAD0Xn              : in std_logic;       -- Counter input for the Paddle 0X.
        PAD0Yn              : in std_logic;       -- Counter input for the Paddle 0Y.
        PAD1Xn              : in std_logic;       -- Counter input for the Paddle 1X.
        PAD1Yn              : in std_logic;       -- Counter input for the Paddle 1Y.
        PADRSTn             : out std_logic;      -- Paddle monoflops reset.

        -- Onewire device:
        CSn_18B20           : out std_logic; -- Onewire device.

        -- Microphone:
        CSn_MP34DB01        : out std_logic
    );
end entity COMBEL_TOP;
    
architecture STRUCTURE of COMBEL_TOP is
signal BERR_In                  : std_logic;
signal ADR_IN_I                 : std_logic_vector(31 downto 1);
signal ASn_I                    : std_logic;
signal ASn_BLT                  : std_logic;
signal BUS_EN_BLT               : std_logic;
signal CSn_18B20_I              : std_logic;
signal DATA_OUT_BLT             : std_logic_vector(15 downto 0);
signal DATA_EN_BLT              : std_logic;
signal DATA_OUT_MCU             : std_logic_vector(15 downto 0);
signal DATA_EN_MCU              : std_logic;
signal DATA_OUT_JOYPORT         : std_logic_vector(15 downto 0);
signal DATA_EN_JOYPORT          : std_logic;
signal DTACK_OUT_IDEn           : std_logic;
signal DTACK_MCUn               : std_logic;
signal DTACK_OUT_BLTn           : std_logic;
signal FC_I                     : std_logic_vector(2 downto 0);
signal FC_BLT                   : std_logic_vector(2 downto 0);
signal ROM_6_In                 : std_logic;
signal ROM_5_In                 : std_logic;
signal ROM_4_In                 : std_logic;
signal ROM_3_In                 : std_logic;
signal ROM_2_In                 : std_logic;
signal ROM_1_In                 : std_logic;
signal ROM_0_In                 : std_logic;
signal SNDCS_In                 : std_logic;
signal DE_I                     : std_logic;
signal KHz_500_I                : std_logic;
signal RTCCS_I                  : std_logic;
signal JOY_RS_I                 : std_logic;
signal PAD0X_RS_I               : std_logic;
signal PAD0Y_RS_I               : std_logic;
signal PAD1X_RS_I               : std_logic;
signal PAD1Y_RS_I               : std_logic;
signal BUTTON_RS_I              : std_logic;
signal SCCn                     : std_logic;
signal VCS_I                    : std_logic;
signal LDS_In                   : std_logic;
signal LDSn_BLT                 : std_logic;
signal MEM_CONFIG_RS_I          : std_logic;
signal LINE_OFFS_RS_I           : std_logic;
signal LINE_WIDTH_RS_I          : std_logic;
signal VIDEO_BASE_HI_RS_I       : std_logic;
signal VIDEO_BASE_MID_RS_I      : std_logic;
signal VIDEO_BASE_LOW_RS_I      : std_logic;
signal VIDEO_COUNT_HI_RS_I      : std_logic;
signal VIDEO_COUNT_MID_RS_I     : std_logic;
signal VIDEO_COUNT_LOW_RS_I     : std_logic;
signal R8006CS_I                : std_logic;
signal R8007CS_I                : std_logic;
signal FPUCS_I                  : std_logic;
signal RAM_In                   : std_logic;
signal R8007_REG                : std_logic_vector(7 downto 0);
signal RWn_I                    : std_logic;
signal RWn_BLT                  : std_logic;
signal SHMOD_ST_SHADOW_RS_I     : std_logic;
signal UDS_In                   : std_logic;
signal UDSn_BLT                 : std_logic;
signal VMODE_SHADOW_RS_I        : std_logic;
signal VIDEO_BASE_HIWORD_RS_I   : std_logic;
signal VIDEO_BASE_LOWORD_RS_I   : std_logic;
signal VIDEO_COUNT_HIWORD_RS_I  : std_logic;
signal VIDEO_COUNT_LOWORD_RS_I  : std_logic;
begin
    ADR_IN_I <= ADR_IN when ADR_16MB_UP = '1' else x"FF" & ADR_IN(23 downto 1);
    ASn_I <= ASn_BLT when BUS_EN_BLT = '1' else 
-- Das sperrt den Interrupthandler.
'1' when FC_I = "111" else ASn; -- No access in the CPU space.
    RWn_I <= RWn_BLT when BUS_EN_BLT = '1' else RWn;
    FC_I <= FC_BLT when BUS_EN_BLT = '1' else FC;

    UDS_In <= UDSn_BLT when BUS_EN_BLT = '1' else UDSn;
    LDS_In <= LDSn_BLT when BUS_EN_BLT = '1' else LDSn;
    
    DATA_EN <= DATA_EN_BLT or DATA_EN_MCU or DATA_EN_JOYPORT;
     -- x"FF" due to pull up resistors in original hardware.
    DATA_OUT <= DATA_OUT_BLT when DATA_EN_BLT = '1' else
                DATA_OUT_MCU when DATA_EN_MCU = '1' else
                DATA_OUT_JOYPORT when DATA_EN_JOYPORT = '1' else 
                x"FF" & DATA_IN(15 downto 8) when R8006CS_I = '1' and RWn_I = '1' else
                x"FF" & R8007_REG when R8007CS_I = '1' and RWn_I = '1' else 
                DATA_IN(15 downto 8) & R8007_REG when R8006CS_I = '1' and R8007CS_I = '1' and RWn_I = '0' else (others => '0');

    -- System controls:
    KHz_500 <= KHz_500_I;
    KHz_500W <= KHz_500_I when TOK = '1' else '0';
    BERRn <= BERR_In;

    --Onewire device:
    CSn_18B20 <= CSn_18B20_I;
    
    -- Register selections:
    R8006 <= R8006CS_I; -- Flacon's configuration register.
    FPUCS <= FPUCS_I; -- Floating point unit.
    VCS <= VCS_I; -- VIDEL video chip.

    -- TOS-ROMS:
    ROM_6n <= ROM_6_In;
    ROM_5n <= ROM_5_In;
    ROM_4n <= ROM_4_In;
    ROM_3n <= ROM_3_In;
    ROM_2n <= ROM_2_In;
    ROM_1n <= ROM_1_In;
    ROM_0n <= ROM_0_In;
    
    -- Soundchip:
    SNDCS   <= '1' when SNDCS_In = '0' and ADR_IN_I(1) = '0' else '0';
    SNDIR   <= '1' when RWn_I = '0' and SNDCS_In = '0' else '0';

    -- Serial communication controller:
    SCCRD <= '1' when SCCn = '0' and RWn_I = '1' else '0';
    SCCWR <= '1' when SCCn = '0' and RWn_I = '0' else '0';

    -- Real time clock:
    RTCCS   <= RTCCS_I;
    RTCAS <= '1' when RTCCS_I = '1' and ADR_IN_I(1) = '0' else '0'; -- x"8961" is the address strobe. 
    RTCDS <= '1' when RTCCS_I = '1' and ADR_IN_I(1) = '1' else '0'; -- x"8963" is the data strobe.
    
    -- Joyport:
    JOY_RHn     <= '0' when JOY_RS_I = '1' and RWn_I = '1' and UDS_In = '0' else '1'; 
    JOY_RLn     <= '0' when JOY_RS_I = '1' and RWn_I = '1' and LDS_In = '0' else '1'; 
    JOY_WL  <= '1' when JOY_RS_I = '1' and RWn_I = '0' and LDS_In = '0' else '0'; 
    JOY_WEn     <= '1' when JOY_RS_I = '1' and RWn_I = '1' and LDS_In = '0' else '0'; 
    BUTTONn     <= '0' when BUTTON_RS_I = '1' else '1';

    -- Feedback for the buscontrollers:
    DTACK_OUTn <=   '0' when ROM_6_In = '0'                     else
                    '0' when ROM_5_In = '0'                     else
                    '0' when ROM_4_In = '0'                     else
                    '0' when ROM_3_In = '0'                     else
                    '0' when ROM_2_In = '0'                     else
                    '0' when ROM_1_In = '0'                     else
                    '0' when ROM_0_In = '0'                     else
                    '0' when SNDCS_In = '0'                     else
                    '0' when RTCCS_I = '1' and RTC_ACK = '1'    else
                    '0' when SCCn = '0' and SCCWAITn = '1'      else
                    '0' when JOY_RS_I = '1'                     else
                    '0' when PAD0X_RS_I = '1'                   else
                    '0' when PAD0Y_RS_I = '1'                   else
                    '0' when PAD1X_RS_I = '1'                   else
                    '0' when PAD1Y_RS_I = '1'                   else
                    '0' when BUTTON_RS_I = '1'                  else
                    '0' when CSn_18B20_I = '0'                  else
                    '0' when R8006CS_I = '1'                    else
                    '0' when R8007CS_I = '1'                    else
                    '0' when FPUCS_I = '1'                      else
                    '0' when VCS_I = '1'                        else
                    '0' when LINE_OFFS_RS_I = '1'               else
                    '0' when LINE_WIDTH_RS_I = '1'              else
                    '0' when MEM_CONFIG_RS_I = '1'              else
                    '0' when VIDEO_BASE_HI_RS_I = '1'           else
                    '0' when VIDEO_BASE_MID_RS_I = '1'          else
                    '0' when VIDEO_BASE_LOW_RS_I = '1'          else
                    '0' when VIDEO_COUNT_HI_RS_I = '1'          else
                    '0' when VIDEO_COUNT_MID_RS_I = '1'         else
                    '0' when VIDEO_COUNT_LOW_RS_I = '1'         else
                    '0' when VIDEO_BASE_HIWORD_RS_I = '1'       else
                    '0' when VIDEO_BASE_LOWORD_RS_I = '1'       else
                    '0' when VIDEO_COUNT_HIWORD_RS_I = '1'      else
                    '0' when VIDEO_COUNT_LOWORD_RS_I = '1'      else
                    '0' when DTACK_OUT_IDEn = '0'               else
                    '0' when DTACK_MCUn = '0'                   else
                    '0' when DTACK_OUT_BLTn = '0'               else '1';

    SYSTEM_REG: process(CLK_32, R8007_REG)
    --    $FF8007|Falcon Bus Control          BIT . . 5 . . 2 . 0 |R/W
    --           |STe Bus Emulation (0 - on) ---------'     |   |
    --           |Blitter (0 - 8mhz, 1 - 16mhz) ------------'   |
    --           |68030 (0 - 8mhz, 1 - 16mhz) ------------------'
    begin
        if CLK_32 = '1' and CLK_32' event then
            if RESET = '1' then
                R8007_REG <= x"00";
            elsif R8007CS_I = '1' and RWn_I = '0' then
                R8007_REG <= DATA_IN(7 downto 0);
            end if;
        end if;
        --
        CPU_CLKSEL_16_8n <= R8007_REG(0);
        BLT_CLKSEL_16_8n <= R8007_REG(2);
    end process SYSTEM_REG;

    I_AUXCLOCKS: CLOCKS
    port map(
        CLK                 => CLK_32,

        CLK_04              => CLK_8,
        CLK_08              => CLK_4,
        CLK_064             => KHz_500_I
    );

    I_BLT: BLITTER_TOP
        port map(
        CLK                 => CLK_BLT,
        RESET               => RESET,
        AS_INn              => ASn_I,
        AS_OUTn             => ASn_BLT,
        LDS_INn             => LDS_In,
        LDS_OUTn            => LDSn_BLT,
        UDS_INn             => UDS_In,
        UDS_OUTn            => UDSn_BLT,
        RWn_IN              => RWn_I,
        RWn_OUT             => RWn_BLT,
        DTACK_INn           => DTACK_INn,
        DTACK_OUTn          => DTACK_OUT_BLTn,
        BERRn               => BERR_In,
        BMODE               => BMODE,
        FC_IN               => FC_I,
        FC_OUT              => FC_BLT,
        BUSCTRL_EN          => BUS_EN_BLT,
        INTn                => BINTn,

        ADR_IN              => ADR_IN_I,
        ADR_OUT             => ADR_OUT,
        ADR_EN              => ADR_EN,
        DATA_IN             => DATA_IN,
        DATA_OUT            => DATA_OUT_BLT,
        DATA_EN             => DATA_EN_BLT,

        BGIn                => BGIn,
        BRn                 => BRn,
        BGACK_INn           => '1',
        BGACK_OUTn          => BGAn,
        BGOn                => BGOn
        );

    I_ADRDEC: ADRDEC
    port map(
        ADR                     => ADR_IN_I,
        RWn                     => RWn_I,

        ROMSEL_FC_E0n           => ROMSEL_FC_E0n,

        LDSn                    => LDS_In,
        UDSn                    => UDS_In,
                
        ASn                     => ASn_I,
        
        VPAn                    => VPAn,
        VMAn                    => VMAn,

        FC                      => FC_I,

        ROM_0n                  => ROM_0_In,
        ROM_1n                  => ROM_1_In,
        ROM_2n                  => ROM_2_In,
        ROM_3n                  => ROM_3_In,
        ROM_4n                  => ROM_4_In,
        ROM_5n                  => ROM_5_In,
        ROM_6n                  => ROM_6_In,

        ACIACS                  => N6850,
        MFPCSn                  => MFPCSn,
        SNDCSn                  => SNDCS_In,
        SCCn                    => SCCn,
        SCCAB                   => SCCAB,
        RTCCS                   => RTCCS_I,

        JOY_RS                  => JOY_RS_I,

        PAD0X_RS                => PAD0X_RS_I,
        PAD0Y_RS                => PAD0Y_RS_I,
        PAD1X_RS                => PAD1X_RS_I,
        PAD1Y_RS                => PAD1Y_RS_I,

        BUTTON_RS               => BUTTON_RS_I,

        CSn_18B20               => CSn_18B20_I,
        CSn_MP34DB01            => CSn_MP34DB01,

        R8006CS                 => R8006CS_I,
        R8007CS                 => R8007CS_I,

        FPUCS                   => FPUCS_I,

        VCS                     => VCS_I,

        MEM_CONFIG_RS           => MEM_CONFIG_RS_I,
        LINE_OFFS_RS            => LINE_OFFS_RS_I,
        LINE_WIDTH_RS           => LINE_WIDTH_RS_I,

        VIDEO_BASE_HIWORD_RS    => VIDEO_BASE_HIWORD_RS_I,
        VIDEO_BASE_LOWORD_RS    => VIDEO_BASE_LOWORD_RS_I,
        VIDEO_COUNT_HIWORD_RS   => VIDEO_COUNT_HIWORD_RS_I,
        VIDEO_COUNT_LOWORD_RS   => VIDEO_COUNT_LOWORD_RS_I,
        VIDEO_BASE_HI_RS        => VIDEO_BASE_HI_RS_I,
        VIDEO_BASE_MID_RS       => VIDEO_BASE_MID_RS_I,
        VIDEO_BASE_LOW_RS       => VIDEO_BASE_LOW_RS_I,
        VIDEO_COUNT_HI_RS       => VIDEO_COUNT_HI_RS_I,
        VIDEO_COUNT_MID_RS      => VIDEO_COUNT_MID_RS_I,
        VIDEO_COUNT_LOW_RS      => VIDEO_COUNT_LOW_RS_I,

        SHMOD_ST_SHADOW_RS      => SHMOD_ST_SHADOW_RS_I,
        VMODE_SHADOW_RS         => VMODE_SHADOW_RS_I,

        RAMn                    => RAM_In
    );

    I_MCU: MCU_TOP
    port map(  
        CLK                     => CLK_32,

        RESET                   => RESET,

        LDSn                    => LDS_In,
        UDSn                    => UDS_In,
        RWn                     => RWn_I,

        ADR_IN                  => ADR_IN_I,
        
        RAMn                    => RAM_In,
        RAMH                    => RAMH,
        VREQ                    => VREQ,
        EVENn_ODD               => EVENn_ODD,

        RDAT                    => RDAT,
        WDAT                    => WDAT,
        VLD                     => VLD,

        VINTn                   => VINTn,
        HINTn                   => HINTn,

        VIDEO_BASE_HIWORD_RS    => VIDEO_BASE_HIWORD_RS_I,
        VIDEO_BASE_LOWORD_RS    => VIDEO_BASE_LOWORD_RS_I,
        VIDEO_COUNT_HIWORD_RS   => VIDEO_COUNT_HIWORD_RS_I,
        VIDEO_COUNT_LOWORD_RS   => VIDEO_COUNT_LOWORD_RS_I,
        VIDEO_BASE_HI_RS        => VIDEO_BASE_HI_RS_I,
        VIDEO_BASE_MID_RS       => VIDEO_BASE_MID_RS_I,
        VIDEO_BASE_LOW_RS       => VIDEO_BASE_LOW_RS_I,
        VIDEO_COUNT_HI_RS       => VIDEO_COUNT_HI_RS_I,
        VIDEO_COUNT_MID_RS      => VIDEO_COUNT_MID_RS_I,
        VIDEO_COUNT_LOW_RS      => VIDEO_COUNT_LOW_RS_I,

        SHMOD_ST_SHADOW_RS      => SHMOD_ST_SHADOW_RS_I,
        VMODE_SHADOW_RS         => VMODE_SHADOW_RS_I,
        MEM_CONFIG_RS           => MEM_CONFIG_RS_I,
        LINE_OFFS_RS            => LINE_OFFS_RS_I,
        LINE_WIDTH_RS           => LINE_WIDTH_RS_I,

        DTACKn                  => DTACK_MCUn,

        DATA_IN                 => DATA_IN,
        DATA_OUT                => DATA_OUT_MCU,
        DATA_EN                 => DATA_EN_MCU,

        RAS1n                   => RAS1n,
        CAS1Hn                  => CAS1Hn,
        CAS1Ln                  => CAS1Ln,
        RAS0n                   => RAS0n,
        CAS0Hn                  => CAS0Hn,
        CAS0Ln                  => CAS0Ln,
        WEn                     => WEn,
        RAM_MADR                => RAM_ADR
    );

    I_INTERRUPT: INTERRUPTS
    port map(
        RESET               => RESET,
        CLK                 => CLK_32,
        ADR_HI              => ADR_IN_I(19 downto 16),
        ADR_LO              => ADR_IN_I(3 downto 1),
        FC                  => FC_I,
        ASn                 => ASn, -- Use Combel input and not ASn_I.
        EINT1n              => EINT1n,
        EINT3n              => EINT3n,
        EINT5n              => EINT5n,
        EINT7n              => EINT7n,
        MFPINTn             => MFPINTn,
        HINTn               => HINTn,
        VINTn               => VINTn,
        AVECn               => AVECn,
        IACKn               => IACKn,
        SCCIACKn            => SCCIACKn,
        IPLn                => IPLn
    );

    I_ErrorHandler: ERRHANDLER
    port map(
        RESET               => RESET,
        CLK                 => CLK_32,
        ASn                 => ASn_I,
        BERRn               => BERR_In
    );

    I_JOYPORT: JOYPORT_PADDLES
    port map(
        CLK                 => CLK_32,
        KHz_500             => KHz_500_I,
        RESET               => RESET,
        
        RWn                 => RWn_I,
        DATA_IN             => DATA_IN(1 downto 0),
        DATA_OUT            => DATA_OUT_JOYPORT,
        DATA_EN             => DATA_EN_JOYPORT,

        PAD0X_RS            => PAD0X_RS_I,
        PAD0Y_RS            => PAD0Y_RS_I,
        PAD1X_RS            => PAD1X_RS_I,
        PAD1Y_RS            => PAD1Y_RS_I,
        
        PAD0X_INHn          => PAD0Xn,
        PAD0Y_INHn          => PAD0Yn,
        PAD1X_INHn          => PAD1Xn,
        PAD1Y_INHn          => PAD1Yn,
        PADRSTn             => PADRSTn
    );

    I_IDE: WF_IDE
    port map(
        RESET               => RESET,

        ADR                 => ADR_IN_I(31 downto 4),

        ASn                 => ASn_I,
        LDSn                => LDS_In,
        RWn                 => RWn_I,
        DTACKn              => DTACK_OUT_IDEn,
        
        -- Interrupt:
        -- HDINTn           =>, -- Not used so far.
        
        -- IDE section:
        IDE_INTRQ           => '0', -- Not used.
        IDE_IORDY           => '1', -- Not used.
        -- PDIAG            =>, -- Not used so far.
        -- DASP             =>, -- See pinout above.
        -- DMARQ            =>, -- Not used so far.
        -- DMACKn           =>, -- See pinout above.
        -- IDE_RESn         =>, -- Not used.
        CS0n                => IDE_RS0n,
        CS1n                => IDE_RS1n,
        IORDn               => IDE_IORDn,
        IOWRn               => IDE_IOWRn,
        
        IDE_D_EN_INn        => IDE_D_EN_INn,
        IDE_D_EN_OUTn       => IDE_D_EN_OUTn
      );

end STRUCTURE;
