----------------------------------------------------------------------
----                                                              ----
---- SD-RAM memory control unit (MCU).				              ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- This MCU is an enhanced version of the original ATARI MCU to ----
---- meet the requirements for modern SD-RAM chips. In detail,    ----
---- this MCU is well suited for the Micron device MT48LC16M16.   ----
---- This controller is systemside compatible to the original     ----
---- Atari MCU CO25912 enhanced with the sound features of STE    ----
---- machines.                                                    ----
----                                                              ----
---- Important Notice concerning the clock system:                ----
---- To use this code in a stand alone MCU chip or in a system    ----
---- on a programmable chip (SOC), the clock frequency must be    ----
---- 16MHz to meet the requirements for the original STs screen   ----
---- resolutions.                                                 ----
---- Affected by the clock selection is the video timing and the  ----
---- DMA sound module (originally in the STE machines).           ----
----            


----- MMMICRON Hinweise und Erlaeuterungen


                                                  ----
---- To guarantee proper operation of the DMA interchange between ----
---- MCU, GLUE, DMA, the clocks must be well selected. For more   ----
---- information see the Suska top level file for the SOC system  ----
---- or respective documentation for the different original types ----
---- of ST or STE machines.                                       ----
----                                                              ----
----                                                              ----
---- To Do:                                                       ----
---- -                                                            ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright � 2009-2010 Wolfgang Foerster Inventronik GmbH.    ----
---- All rights reserved. No portion of this sourcecode may be    ----
---- reproduced or transmitted in any form by any means, whether  ----
---- by electronic, mechanical, photocopying, recording or        ----
---- otherwise, without my written permission.                    ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K9B  2009/12/24 WF
--   Initial release.

library work;
use work.COMBEL_PKG.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity MCU_TOP is
	port(  
		CLK				        : in std_logic; -- System clock, 32 MHz.

		RESET   		        : in std_logic;
    
		LDSn, UDSn		        : in std_logic; -- Bus control signals.
		RWn				        : in std_logic; -- Bus control signals.

		ADR_IN		            : in std_logic_vector(31 downto 1); -- The Falcon address bus.

        RAMn			        : in std_logic;
        RAMH			        : out std_logic; -- Hi bank select signal.
        VREQ                    : in std_logic;
        EVENn_ODD               : in std_logic;
		
		RDAT			        : out std_logic; -- Buffer control.
		WDAT			        : out std_logic; -- Buffer control.
		VLD 			        : out std_logic; -- Buffer control.
			
        VINTn			        : in std_logic; -- Vertical sync.
        HINTn			        : in std_logic; -- Horizontal sync.
            
        VIDEO_BASE_HIWORD_RS	: in std_logic;
        VIDEO_BASE_LOWORD_RS	: in std_logic;
        VIDEO_COUNT_HIWORD_RS	: in std_logic;
        VIDEO_COUNT_LOWORD_RS	: in std_logic;

        VIDEO_BASE_HI_RS	    : in std_logic;
        VIDEO_BASE_MID_RS	    : in std_logic;
        VIDEO_BASE_LOW_RS	    : in std_logic;
        VIDEO_COUNT_HI_RS	    : in std_logic;
        VIDEO_COUNT_MID_RS	    : in std_logic;
        VIDEO_COUNT_LOW_RS	    : in std_logic;
        SHMOD_ST_SHADOW_RS      : in std_logic;
        VMODE_SHADOW_RS         : in std_logic;
        MEM_CONFIG_RS		    : in std_logic;
        LINE_OFFS_RS		    : in std_logic;
        LINE_WIDTH_RS		    : in std_logic;

		DTACKn			        : out std_logic; -- Data acknowledge signal.

		DATA_IN			        : in std_logic_vector(15 downto 0);
		DATA_OUT		        : out std_logic_vector(15 downto 0);
		DATA_EN			        : out std_logic;

        -- RAM interface:
        RAS1n			        : out std_logic;
        CAS1Hn  		        : out std_logic;
        CAS1Ln  		        : out std_logic;
        RAS0n			        : out std_logic;
        CAS0Hn			        : out std_logic;
        CAS0Ln  		        : out std_logic;
        WEn				        : out std_logic;

        RAM_MADR			    : out std_logic_vector(26 downto 1)
	);
end entity MCU_TOP;

architecture STRUCTURE of MCU_TOP is
signal RAS1_CTRLn			: std_logic;
signal CAS1H_CTRLn			: std_logic;
signal CAS1L_CTRLn			: std_logic;
signal RAS0_CTRLn			: std_logic;
signal CAS0H_CTRLn			: std_logic;
signal CAS0L_CTRLn			: std_logic;
signal WE_CTRLn				: std_logic;

signal DATA_OUT_CTRL		: std_logic_vector(7 downto 0);
signal DATA_EN_CTRL			: std_logic;
signal DATA_OUT_VCNT		: std_logic_vector(15 downto 0);
signal DATA_EN_VCNT			: std_logic;

signal VIDEO_CNT_EN_I		: std_logic;
signal VIDEO_CNT_LOAD_I		: std_logic;

signal M_ADR_I				: std_logic_vector(31 downto 1);
signal VIDEO_ADR_I			: std_logic_vector(31 downto 1);
signal MCU_PHASE_I			: MCU_PHASE_TYPE;
signal REF_EN_I				: std_logic;

begin
    RAM_MADR <= M_ADR_I(26 downto 1); -- Fit this to the used RAM.
	M_ADR_I <= ADR_IN when MCU_PHASE_I = RAM else
               VIDEO_ADR_I when MCU_PHASE_I = VIDEO else (others => '0'); -- Video is in the lower 16MB.

    RAS0n <= '0' when REF_EN_I = '1' else RAS0_CTRLn;
    RAS1n <= '0' when REF_EN_I = '1' else RAS1_CTRLn;
    CAS1Hn <= '0' when REF_EN_I = '1' else CAS1H_CTRLn;
    CAS1Ln <= '0' when REF_EN_I = '1' else CAS1L_CTRLn;
    CAS0Hn <= '0' when REF_EN_I = '1' else CAS0H_CTRLn;
    CAS0Ln <= '0' when REF_EN_I = '1' else CAS0L_CTRLn;
    WEn <= WE_CTRLn;

	DATA_EN <= DATA_EN_CTRL or DATA_EN_VCNT;
    DATA_OUT <= x"00" & DATA_OUT_CTRL when DATA_EN_CTRL = '1' else
				DATA_OUT_VCNT when DATA_EN_VCNT = '1' else (others => '0');

	I_CONTROL: MCU_CTRL
		port map(
			CLK 			=> CLK,
			RESET  			=> RESET,

			LDSn 			=> LDSn,
			UDSn 			=> UDSn,
			RWn 			=> RWn,
			
			M_ADR 			=> M_ADR_I,
            
			RAMn 			=> RAMn,
            RAMH            => RAMH,
            VREQ            => VREQ,

			MEM_CONFIG_RS 	=> MEM_CONFIG_RS,

			VINTn 			=> VINTn,

			RAS0n 			=> RAS0_CTRLn,
			CAS0Hn 			=> CAS0H_CTRLn,
			CAS0Ln 			=> CAS0L_CTRLn,

			WEn 			=> WE_CTRLn,

			RAS1n 			=> RAS1_CTRLn,
			CAS1Hn 			=> CAS1H_CTRLn,
			CAS1Ln 			=> CAS1L_CTRLn,

			RDAT  			=> RDAT,
			WDAT  			=> WDAT,
			VLD  			=> VLD,
			
			REF_EN 			=> REF_EN_I,
			VIDEO_CNT_EN 	=> VIDEO_CNT_EN_I,
			VIDEO_CNT_LOAD 	=> VIDEO_CNT_LOAD_I,
			
			MCU_PHASE 		=> MCU_PHASE_I,
			
			DTACKn			=> DTACKn,

			DATA_IN			=> DATA_IN(7 downto 0),
			DATA_OUT		=> DATA_OUT_CTRL,
			DATA_EN			=> DATA_EN_CTRL
		);

	I_VIDEO: MCU_VIDEO_COUNTER
		port map(
			CLK		                => CLK,
			RESET  		            => RESET,

			RWn			            => RWn,

			VIDEO_BASE_HIWORD_RS 	=> VIDEO_BASE_HIWORD_RS,
			VIDEO_BASE_LOWORD_RS 	=> VIDEO_BASE_LOWORD_RS,
			VIDEO_COUNT_HIWORD_RS 	=> VIDEO_COUNT_HIWORD_RS,
			VIDEO_COUNT_LOWORD_RS 	=> VIDEO_COUNT_LOWORD_RS,
		
			VIDEO_BASE_HI_RS 	    => VIDEO_BASE_HI_RS,
			VIDEO_BASE_MID_RS 	    => VIDEO_BASE_MID_RS,
			VIDEO_BASE_LOW_RS 	    => VIDEO_BASE_LOW_RS,

			VIDEO_COUNT_HI_RS 	    => VIDEO_COUNT_HI_RS,
			VIDEO_COUNT_MID_RS 	    => VIDEO_COUNT_MID_RS,
			VIDEO_COUNT_LOW_RS 	    => VIDEO_COUNT_LOW_RS,
		
            SHMOD_ST_SHADOW_RS      => SHMOD_ST_SHADOW_RS,
            VMODE_SHADOW_RS         => VMODE_SHADOW_RS,

            VINTn                   => VINTn,
            HINTn                   => HINTn,

			EVENn_ODD               => EVENn_ODD,
			VIDEO_COUNT_EN 		    => VIDEO_CNT_EN_I,
			VIDEO_COUNT_LOAD 	    => VIDEO_CNT_LOAD_I,

			LINE_OFFS_RS 		    => LINE_OFFS_RS,
			LINE_WIDTH_RS 		    => LINE_WIDTH_RS,
		
			VIDEO_ADR_OUT		    => VIDEO_ADR_I,
		
			DATA_IN		            => DATA_IN,
			DATA_OUT	            => DATA_OUT_VCNT,
			DATA_EN		            => DATA_EN_VCNT
      	);

end architecture STRUCTURE;
