---------------------------------------------------------------------
----                                                             ----
----  Serial EEPROM master controller.        				     ----
----                                                             ----
----  Author: Jens Carroll, Wolfgang Foerster                    ----
----          support@inventronk.de                              ----
----          www.inventronik.de                                 ----
----                                                             ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright � 2009-2012 Wolfgang Foerster Inventronik GmbH.    ----
---- All rights reserved. No portion of this sourcecode may be    ----
---- reproduced or transmitted in any form by any means, whether  ----
---- by electronic, mechanical, photocopying, recording or        ----
---- otherwise, without my written permission.                    ----
----                                                              ----
----------------------------------------------------------------------
----															 ----
---- Description:												 ----
---- This module is a controller for I2C controlled EEPROMS of   ----
---- the 24xxx fanmily. It is suited to work in conjunction with ----
---- a state machine controlling the data flow.                  ----
----                                                             ----
---- The logic is controlled by the following signals:           ----
----    RD_RND, RD_SEQ, WR_BYTE, WR_PAGE which are inputs.       ----
----    RDY, REQ which are outputs.                              ----
----    The 24xxx controller knows four modi:                    ----
----    1. The read random mode. In this mode the EEPROM is read ----
----       randomly. Refer to the respective date sheets for     ----
----       more information. To initiate this operation, the     ----
----       signal RD_RND must be asserted and deasserted when    ----
----       the REQ_RDY signal is asserted by the 24xxx control-  ----
----       ler. The data to be read is valid when the RDY signal ----
----       is asserted. The data to be read is valid after the   ----
----       REQ_RDY signal is asserted by the 24xxx controller    ----
----       until the next operation.                             ----
----       Be aware, that REQ_RDY is a strobe respective to CLK. ----
----    2. The write byte mode. In this mode the EEPROM is writ- ----
----       ten byte wise. Refer to the respective date sheets    ----
----       for more information. To initiate this operation, the ----
----       signal WR_BYTE must be asserted and deasserted when   ----
----       the REQ_RDY signal is asserted by the 24xxx control-  ----
----       ler. The data to be written must be valid form asser- ----
----       ting WR_BYTE until the REQ_RDY signal is asserted.    ----
----       Be aware, that REQ_RDY is a strobe respective to CLK. ----
----    3. The read sequential mode. In this mode the EEPROM is  ----
----       read out sequentially. Refer to the respective date   ----
----       sheets for more information. To initiate this opera-  ----
----       tion, the signal RD_SEQ must be asserted. During this ----
----       operation the 24xxx controller indicates via the      ----
----       REQ_RDY signal, that new data is ready to read from   ----
----       the DATA_OUT port. The data should be read from the   ----
----       port immediately after this signal is asserted. The   ----
----       read process takes place as long as the RD_SEQ is     ----
----       asserted. To finish this mode, the RD_SEQ must be     ----
----       deasserted during REQ_RDY is asserted for the last    ----
----       byte to be read. Be aware, that the read sequential   ----
----       mode is alway started by writing the control byte and ----
----       the base address.                                     ----
----       Be aware, that REQ_RDY is a strobe respective to CLK. ----
----    4. The write page mode. In this mode, 64 or 128 bytes    ----
----       are written sequentially depending of the chip type.  ----
----       See the generic setting for adjusting this correctly. ----
----       Refer to the respective date sheets for more infor-   ----
----       mation. To initiate this operation, the signal        ----
----       WR_PAGE must be asserted. During this operation the   ----
----       REQ_RDY signal indicates, that new data on the        ----
----       DATA_IN port is required to be written. The write     ----
----       page mode is active as long as the signal WR_PAGE is  ----
----       asserted. A good way to finisch this mode is to dis-  ----
----       able the WR_PAGE signal after the last REQ_RDY and    ----
----       after a complete page is written. This means, that    ----
----       the master controller which is responsible for asser- ----
----       ting the WR_PAGE signal must determine the amount of  ----
----       bytes written by counting the REQ_RDY strobes.        ----
----       Be aware, that REQ_RDY is a strobe respective to CLK. ----
---------------------------------------------------------------------

-- Change History:
-- Revision 1.0  20090720, w. foerster.
--   Initial release.
-- Revision 1.1  20120222, w. foerster.
--   Improvements concerning page write and sequential read.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity EEPROM_24xxx is
generic(I2C_ADR         : bit_vector(2 downto 0) := "000"; -- Hardwired.
        I2C_CTRL_CODE   : bit_vector(3 downto 0) := "1010"; -- The control code matches the 24xxx.
        --BURST_WIDTH   : bit_vector(7 downto 0) := x"40"); -- Use this for 24256 EEPROM.
        BURST_WIDTH     : std_logic_vector(7 downto 0) := x"80"); -- Use this for 24512 EEPROM.
port(
    CLK			: in bit;
    RESET		: in bit;

    ADR_IN      : in bit_vector(15 downto 0); -- The MSB is not applicable for the 24LC256 (don't care).
    DATA_IN	    : in bit_vector(7 downto 0);
    DATA_OUT	: out bit_vector (7 downto 0);

    RD_RND      : in bit;
    RD_SEQ      : in bit;
    WR_BYTE     : in bit;
    WR_PAGE     : in bit;

    REQ_RDY     : out bit;
    
    I2C_SDA_IN  : in  bit;
    I2C_SDA_OUT : out  bit;
    I2C_SDA_EN  : out bit;
    I2C_SCL     : out  bit;
    I2C_SCL_EN  : out  bit
    );
end entity EEPROM_24xxx;

architecture BEHAVIOR of EEPROM_24xxx is
type EEPROM_CTRLTYPE is (IDLE, START_11, START_12, START_13, START_21, START_22, START_23, WAIT_ACK_1, WAIT_ACK_2,
                         WAIT_ACK_3, WAIT_ACK_4, WAIT_ACK_5, SND_ACK, SND_NACK, CONTROL_1, CONTROL_2, ADDRESS_HI,
                         ADDRESS_LO, SND_DATA, RCV_DATA, STOP_11, STOP_12, STOP_13);
signal EEPROM_CTRL      : EEPROM_CTRLTYPE;
signal NEXT_EEPROM_CTRL : EEPROM_CTRLTYPE;
signal I2C_STRB         : bit;
signal BYTE_RDY         : bit;
signal PAGE_RDY         : bit;
signal SCL_OUT          : bit;
signal SDA_IN           : bit;
signal SDA_OUT          : bit;
begin
    TIMEBASE: process(CLK) -- CLK is 70MHz.
    -- This logic provides an adequate clock for the I2C device
    -- of about 300kHz. Adjust the SCL clock in this process. 
    -- Example: Input frequency is 20MHz, the output is 300kHz. 
    -- The divider is 70.000.000 / 300.000 = 234. Chose TMP
    -- to >= 7 bit. Adjust the TMP rollover to a value of 234.
    variable TMP    : std_logic_vector(7 downto 0);
    begin
        if CLK = '1' and CLK' event then
            if RESET = '1' or TMP = x"00" then
                TMP := x"EA";
            else
                TMP := TMP - '1';
            end if;
            --
            case TMP is
                when x"00" => I2C_STRB <= '1';
                when others => I2C_STRB <= '0';
            end case;
            --
            -- The data is aligned to be stable during the rising 
            -- or the falling edge of SCL -> SCL is shifted versus
            -- I2C_STRB.
            if TMP > x"3A" and TMP < x"AF" then
                SCL_OUT <= '1';
            else
                SCL_OUT <= '0';
            end if;
        end if;
    end process TIMEBASE;
    
    I2C_SDA_EN <= '1' when EEPROM_CTRL = START_11 or EEPROM_CTRL = START_12 or EEPROM_CTRL = START_13 else
                  '1' when EEPROM_CTRL = START_21 or EEPROM_CTRL = START_22 or EEPROM_CTRL = START_23 else
                  '1' when EEPROM_CTRL = CONTROL_1 or EEPROM_CTRL = CONTROL_2 else
                  '1' when EEPROM_CTRL = ADDRESS_HI or EEPROM_CTRL = ADDRESS_LO else
                  '1' when EEPROM_CTRL = SND_DATA else
                  '1' when EEPROM_CTRL = SND_ACK or EEPROM_CTRL = SND_NACK else
                  '1' when EEPROM_CTRL = STOP_11 or EEPROM_CTRL = STOP_12 or EEPROM_CTRL = STOP_13 else '0';

    I2C_SDA_OUT <= '1' when EEPROM_CTRL = START_11 or EEPROM_CTRL = START_21 else
                   '0' when EEPROM_CTRL = START_12 or EEPROM_CTRL = START_22 else
                   '0' when EEPROM_CTRL = START_13 or EEPROM_CTRL = START_23 else
                   '0' when EEPROM_CTRL = STOP_11 or EEPROM_CTRL = STOP_12 else
                   '0' when EEPROM_CTRL = SND_ACK else
                   '1' when EEPROM_CTRL = SND_NACK else
                   '1' when EEPROM_CTRL = STOP_13 else SDA_OUT;
    
    I2C_SCL_EN <= '1' when EEPROM_CTRL /= IDLE else '0';
    
    I2C_SCL <= '1' when EEPROM_CTRL = START_11 or EEPROM_CTRL = START_21 else
               '1' when EEPROM_CTRL = START_12 or EEPROM_CTRL = START_22 else
               '0' when EEPROM_CTRL = START_13 or EEPROM_CTRL = START_23 else
               '0' when EEPROM_CTRL = STOP_11 else
               '1' when EEPROM_CTRL = STOP_12 or EEPROM_CTRL = STOP_13 else SCL_OUT;

    REQ_RDY <= '1' when RD_RND = '1' and EEPROM_CTRL /= IDLE and NEXT_EEPROM_CTRL = IDLE and I2C_STRB = '1' else
               '1' when WR_BYTE = '1' and EEPROM_CTRL /= IDLE and NEXT_EEPROM_CTRL = IDLE and I2C_STRB = '1' else
               '1' when RD_SEQ = '1' and (EEPROM_CTRL = SND_ACK or EEPROM_CTRL = SND_NACK) and I2C_STRB = '1' else
               '1' when WR_PAGE = '1' and EEPROM_CTRL = SND_DATA and NEXT_EEPROM_CTRL /= SND_DATA and I2C_STRB = '1' else '0';

    I2C_CTRL_REG: process(CLK)
    -- This is the I2C's state machine register.
    begin
        if CLK = '1' and CLK' event then
            if RESET = '1' then
                EEPROM_CTRL <= IDLE;
            elsif I2C_STRB = '1' then
                EEPROM_CTRL <= NEXT_EEPROM_CTRL;
            end if;
        end if;
    end process I2C_CTRL_REG;
    
    I2C_CTRL_DEC: process(EEPROM_CTRL, RD_SEQ, RD_RND, WR_PAGE, WR_BYTE, BYTE_RDY, PAGE_RDY, SDA_IN)
    -- This is the I2C's state machine decoder.
    begin
        case EEPROM_CTRL is
            when IDLE =>
                if RD_RND = '1' or RD_SEQ = '1' or WR_BYTE = '1' or WR_PAGE = '1' then
                    NEXT_EEPROM_CTRL <= START_11;
                else
                    NEXT_EEPROM_CTRL <= IDLE;
                end if;
            when START_11 =>
                NEXT_EEPROM_CTRL <= START_12;
            when START_12 =>
                NEXT_EEPROM_CTRL <= START_13;
            when START_13 =>
                NEXT_EEPROM_CTRL <= CONTROL_1;
            when CONTROL_1 =>
                if BYTE_RDY = '1' then
                    NEXT_EEPROM_CTRL <= WAIT_ACK_1;
                else
                    NEXT_EEPROM_CTRL <= CONTROL_1;
                end if;
            when WAIT_ACK_1 =>
                if SDA_IN = '0' then
                    NEXT_EEPROM_CTRL <= ADDRESS_HI;
                else
                    NEXT_EEPROM_CTRL <= START_11; -- Device is Busy!
                end if;
            when ADDRESS_HI =>
                if BYTE_RDY = '1' then
                    NEXT_EEPROM_CTRL <= WAIT_ACK_2;
                else
                    NEXT_EEPROM_CTRL <= ADDRESS_HI;
                end if;
            when WAIT_ACK_2 =>
                if SDA_IN = '0' then
                    NEXT_EEPROM_CTRL <= ADDRESS_LO;
                else
                    NEXT_EEPROM_CTRL <= IDLE; -- Error!
                end if;
            when ADDRESS_LO =>
                if BYTE_RDY = '1' then
                    NEXT_EEPROM_CTRL <= WAIT_ACK_3;
                else
                    NEXT_EEPROM_CTRL <= ADDRESS_LO;
                end if;
            when WAIT_ACK_3 =>
                if SDA_IN = '0' and WR_BYTE = '1' then
                    NEXT_EEPROM_CTRL <= SND_DATA;
                elsif SDA_IN = '0' and WR_PAGE = '1' then
                    NEXT_EEPROM_CTRL <= SND_DATA;
                elsif SDA_IN = '0' and RD_RND = '1' then
                    NEXT_EEPROM_CTRL <= START_21;
                elsif SDA_IN = '0' and RD_SEQ = '1' then
                    NEXT_EEPROM_CTRL <= START_21;
                else
                    NEXT_EEPROM_CTRL <= IDLE; -- Error!
                end if;
            when SND_DATA =>
                if BYTE_RDY = '1' then
                    NEXT_EEPROM_CTRL <= WAIT_ACK_4;
                else
                    NEXT_EEPROM_CTRL <= SND_DATA;
                end if;
            when WAIT_ACK_4 =>
                if SDA_IN = '0' and WR_PAGE = '1' and PAGE_RDY = '0' then
                    NEXT_EEPROM_CTRL <= SND_DATA; -- Page write.
                elsif SDA_IN = '0' then -- Byte written.
                    NEXT_EEPROM_CTRL <= STOP_11;
                else
                    NEXT_EEPROM_CTRL <= IDLE; -- Error!
                end if;
            when START_21 =>
                NEXT_EEPROM_CTRL <= START_22;
            when START_22 =>
                NEXT_EEPROM_CTRL <= START_23;
            when START_23 =>
                NEXT_EEPROM_CTRL <= CONTROL_2;
            when CONTROL_2 =>
                if BYTE_RDY = '1' then
                    NEXT_EEPROM_CTRL <= WAIT_ACK_5;
                else
                    NEXT_EEPROM_CTRL <= CONTROL_2;
                end if;
            when WAIT_ACK_5 =>
                if SDA_IN = '0' then
                    NEXT_EEPROM_CTRL <= RCV_DATA;
                else
                    NEXT_EEPROM_CTRL <= IDLE; -- Error!
                end if;
            when RCV_DATA =>
                if BYTE_RDY = '1' and RD_SEQ = '1' then
                    NEXT_EEPROM_CTRL <= SND_ACK;
                elsif BYTE_RDY = '1' then -- Finish.
                    NEXT_EEPROM_CTRL <= SND_NACK;
                else
                    NEXT_EEPROM_CTRL <= RCV_DATA;
                end if;
            when SND_ACK =>
                NEXT_EEPROM_CTRL <= RCV_DATA;
            when SND_NACK =>
                NEXT_EEPROM_CTRL <= STOP_11;
            when STOP_11 =>
                NEXT_EEPROM_CTRL <= STOP_12;
            when STOP_12 =>
                NEXT_EEPROM_CTRL <= STOP_13;
            when STOP_13 =>
                NEXT_EEPROM_CTRL <= IDLE;
        end case;
    end process I2C_CTRL_DEC;

    P_SDA_IN: process(CLK)
    -- This flip flop stores the information of the SDA line
    -- when the SCLK is high.
    variable LOCK    : boolean;
    begin
        if CLK = '1' and CLK' event then
            if RESET = '1' then
                LOCK := false;
                SDA_IN <= '1';
            elsif SCL_OUT = '1' and LOCK = false then
                LOCK := true;
                SDA_IN <= I2C_SDA_IN;
            elsif SCL_OUT = '0' then
                LOCK := false;
            end if;
        end if;
    end process P_SDA_IN;

    BITCOUNT: process(CLK)
    -- This process controls the bits transfered from or to the slave.
    variable BIT_CNT    : integer range 0 to 7;
    begin
        if CLK = '1' and CLK' event then
            if RESET = '1' then
                BIT_CNT := 0;
            elsif EEPROM_CTRL /= CONTROL_1 and NEXT_EEPROM_CTRL = CONTROL_1 then
                BIT_CNT := 0;
            elsif EEPROM_CTRL /= CONTROL_2 and NEXT_EEPROM_CTRL = CONTROL_2 then
                BIT_CNT := 0;
            elsif EEPROM_CTRL /= ADDRESS_HI and NEXT_EEPROM_CTRL = ADDRESS_HI then
                BIT_CNT := 0;
            elsif EEPROM_CTRL /= ADDRESS_LO and NEXT_EEPROM_CTRL = ADDRESS_LO then
                BIT_CNT := 0;
            elsif EEPROM_CTRL /= SND_DATA and NEXT_EEPROM_CTRL = SND_DATA then
                BIT_CNT := 0;
            elsif EEPROM_CTRL /= RCV_DATA and NEXT_EEPROM_CTRL = RCV_DATA then
                BIT_CNT := 0;
            elsif I2C_STRB = '1' and BIT_CNT < 7 then
                BIT_CNT := BIT_CNT + 1;
            end if;
            --
            case BIT_CNT is
                when 7 => BYTE_RDY <= '1';
                when others => BYTE_RDY <= '0';
            end case;
        end if;
    end process BITCOUNT;

    BYTECOUNT: process(CLK)
    -- This process controls the data flow during a page write access. The transfer is ready,
    -- if a page has been written.
    variable BYTE_CNT    : std_logic_vector(7 downto 0);
    begin
        if CLK = '1' and CLK' event then
            if RESET = '1' then
                BYTE_CNT := (others => '0');
            elsif EEPROM_CTRL /= SND_DATA and EEPROM_CTRL /= WAIT_ACK_4 then
                BYTE_CNT := (others => '0');
            elsif I2C_STRB = '1' and EEPROM_CTRL = SND_DATA and BYTE_RDY = '1' then
                BYTE_CNT := BYTE_CNT + '1';            
            end if;
            --
            case BYTE_CNT is
                when BURST_WIDTH => PAGE_RDY <= '1'; -- 1 Page transfered.
                when others => PAGE_RDY <= '0';
            end case;
        end if;
    end process BYTECOUNT;
    
    SHIFTREG: process(CLK)
    variable SHIFT_REG          : bit_vector(7 downto 0);
    begin
        if CLK = '1' and CLK' event then
            if RESET = '1' then
                SHIFT_REG := x"00";
            elsif EEPROM_CTRL /= CONTROL_1 and NEXT_EEPROM_CTRL = CONTROL_1 then -- Load control data.
                SHIFT_REG := I2C_CTRL_CODE & I2C_ADR & '0'; -- Write address (and data for write to slave).
            elsif EEPROM_CTRL /= CONTROL_2 and NEXT_EEPROM_CTRL = CONTROL_2 then -- Load control data.
                SHIFT_REG := I2C_CTRL_CODE & I2C_ADR & '1'; -- Read from slave.
            elsif EEPROM_CTRL /= ADDRESS_HI and NEXT_EEPROM_CTRL = ADDRESS_HI then -- Load data to be sent.
                SHIFT_REG := ADR_IN(15 downto 8);
            elsif EEPROM_CTRL /= ADDRESS_LO and NEXT_EEPROM_CTRL = ADDRESS_LO then -- Load data to be sent.
                SHIFT_REG := ADR_IN(7 downto 0);
            elsif EEPROM_CTRL /= SND_DATA and NEXT_EEPROM_CTRL = SND_DATA then -- Load data to be sent.
                SHIFT_REG := DATA_IN;
            elsif EEPROM_CTRL = CONTROL_1 and I2C_STRB = '1' then -- Shift out.
                SHIFT_REG := SHIFT_REG(6 downto 0) & '0';
            elsif EEPROM_CTRL = CONTROL_2 and I2C_STRB = '1' then -- Shift out.
                SHIFT_REG := SHIFT_REG(6 downto 0) & '0';
            elsif EEPROM_CTRL = ADDRESS_HI and I2C_STRB = '1' then -- Shift out.
                SHIFT_REG := SHIFT_REG(6 downto 0) & '0';
            elsif EEPROM_CTRL = ADDRESS_LO and I2C_STRB = '1' then -- Shift out.
                SHIFT_REG := SHIFT_REG(6 downto 0) & '0';
            elsif EEPROM_CTRL = SND_DATA and I2C_STRB = '1' then -- Shift out.
                SHIFT_REG := SHIFT_REG(6 downto 0) & '0';
            elsif EEPROM_CTRL = RCV_DATA and I2C_STRB = '1' then -- Shift in.
                SHIFT_REG := SHIFT_REG(6 downto 0) & SDA_IN;
            end if;
            --
            if BYTE_RDY = '1' then
                DATA_OUT <= SHIFT_REG;
            end if;
            --
            SDA_OUT <= SHIFT_REG(7);
        end if;
    end process SHIFTREG;
end architecture BEHAVIOR;
