----------------------------------------------------------------------
----                                                              ----
---- ATARI GLUE compatible IP Core					              ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- Atari's ST Glue with all features to reach                   ----
---- ATARI STE compatibility.                                     ----
----                                                              ----
---- Direct memory access (DMA) control state machine.            ----
----                                                              ----
----                                                              ----
---- To Do:                                                       ----
---- -                                                            ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright (C) 2005 - 2011 Wolfgang Foerster                  ----
----                                                              ----
---- This source file may be used and distributed without         ----
---- restriction provided that this copyright statement is not    ----
---- removed from the file and that any derivative work contains  ----
---- the original copyright notice and the associated disclaimer. ----
----                                                              ----
---- This source file is free software; you can redistribute it   ----
---- and/or modify it under the terms of the GNU Lesser General   ----
---- Public License as published by the Free Software Foundation; ----
---- either version 2.1 of the License, or (at your option) any   ----
---- later version.                                               ----
----                                                              ----
---- This source is distributed in the hope that it will be       ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied   ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ----
---- PURPOSE. See the GNU Lesser General Public License for more  ----
---- details.                                                     ----
----                                                              ----
---- You should have received a copy of the GNU Lesser General    ----
---- Public License along with this source; if not, download it   ----
---- from http://www.gnu.org/licenses/lgpl.html                   ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K6A  2006/06/03 WF
--   Initial Release.
-- Revision 2K8A  2008/02/13 WF
--   Upgraded to version V1:
--     DMA_SYNC is now foreseen to synchronize the DMA
--     transfer between MCU, GLUE and DMA.
-- Revision 2K8B  2008/12/24 WF
--   Minor changes concerning DMA_SYNC.
-- Revision 2K9A  2009/06/20 WF
--   BRN_LOGIC process has now synchronous reset to provide preset requirement.
-- Revision 2K9B  2009/12/24 WF
--   Partially rewritten the bus arbiter, removed DMA_SYNC, improved
--     the bus timing for read and write operation.
--   Fixed a bug in the BRn_LOGIC process not to start the DMA operation unintendedly.
-- Revision 2K15B  20151224 WF
--   Replaced the data type bit by std_logic.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity WF25915IP_BUS_ARBITER_V2 is
port (  RESn	        : in std_logic; -- ST's reset signal.
		CLK		        : in std_logic; -- the ST's 8MHz clock.

		-- D8 is used for the DMA mode register.
		D8		        : in std_logic; -- Data bus bit 8.
		
		RAMn            : in std_logic; -- Indicates RAM access.
		DTACKn		    : in std_logic; -- DTACKn gives information about the MMU state.

		-- ASn, LDSn, UDSn and RWn are asserted by the arbiter during 
		-- DMA transfer.
		AS_INn			: in std_logic;
		AS_OUTn			: out std_logic;
		RWn_OUT			: out std_logic;
		LDS_OUTn		: out std_logic;
		UDS_OUTn		: out std_logic;
        FC_OUT          : out std_logic_vector(2 downto 0);
		CTRL_EN			: out std_logic;

		-- Chip select signal for the DMA mode register. This signal
		-- is generated in the adress decoder.
	    DMA_MODE_CSn	: in std_logic;

		-- RDYn is asserted by the arbiter during DMA transfer.
		-- The other signals are used for DMA transfer control.
		RDY_INn		    : in std_logic; -- DMA unit ready signal.
		RDY_OUTn	    : out std_logic; -- DMA unit ready signal.
		BGACK_INn	    : in std_logic; -- Bus grant acknowledge input.
		BGACK_OUTn	    : out std_logic; -- Bus grant acknowledge output.
		BRn			    : out std_logic; -- Bus request (open drain).
		BGIn		    : in std_logic; -- Bus grant input.
		BGOn		    : out std_logic; -- Bus grant output.
		DMAn		    : out std_logic -- DMA select signal for the MMU.
      );
end WF25915IP_BUS_ARBITER_V2;

architecture BEHAVIOR of WF25915IP_BUS_ARBITER_V2 is
type DMA_PHASES is (IDLE, SYNC, READ, WRITE);
type TIME_SLICES is (IDLE, S0, S1, S2, S3, S4, S5, S6, S7);
signal DMA_PHASE		: DMA_PHASES;
signal T_SLICE			: TIME_SLICES;
signal DMA_MODE			: std_logic; -- One bit register.
signal BR_In			: std_logic;
signal SLICE_CNT		: std_logic_vector(2 downto 0);
signal SYNC_CNT_N	    : std_logic_vector(1 downto 0);
signal UDS_WR_EN        : std_logic;
signal UDS_RD_EN        : std_logic;
signal LDS_WR_EN        : std_logic;
signal LDS_RD_EN        : std_logic;
signal RDYn             : std_logic;
begin
	BRn <= BR_In;
	BGOn <= '0' when BGIn = '0' and BR_In = '1' else '1';
	BGACK_OUTn <= '1' when DMA_PHASE = IDLE else '0';

    DMAn <= '1' when DMA_PHASE = IDLE or DMA_PHASE = SYNC else '0';

	DMA_MODE_REG: process(RESn, CLK)
	-- This is the DMA mode register in the GLUE. it is a mirror of the 
	-- original DMA mode register (the original is 8 bit wide), located
	-- in the DMA unit. This register stores the information about the
	-- DMA data transfer direction. The STs adress of this register is
	-- FF8606. The register select signal DMA_MODE_CSn is decoded in the 
	-- adress decoder. This register is write only.
	begin
		if RESn = '0' then
			DMA_MODE <= '0'; -- Default is read from ACSI / Floppy.
        elsif CLK = '1' and CLK' event then
			if DMA_MODE_CSn = '0' then
				DMA_MODE <= D8;
			end if;
		end if;
	end process DMA_MODE_REG;

	BRn_LOGIC: process
    variable RDY_COUNT: std_logic_vector(2 downto 0);
	begin
		wait until CLK = '1' and CLK' event;
		if RESn = '0' then
			BR_In <= '1';
            RDY_COUNT := "000";
        elsif RDY_INn = '1' and RDY_COUNT /= "111" then
			RDY_COUNT := RDY_COUNT + '1'; -- RDYn condition detected.
		elsif RDY_INn = '0' then -- Restart the counter.
            RDY_COUNT := "000";
		end if;
		case RDY_COUNT is
            when "111" | "110" | "101" => BR_In <= '0'; -- DMA condition detected.
			when others => BR_In <= '1';
		end case;
	end process BRn_LOGIC;

	DMA_CTRL: process(CLK, RESn, DMA_PHASE, DTACKn)
    -- State machine for DMA sequence detection. During the DMA sequence, the
    -- signals LDSn, UDSn, ASn, RDYn and RWn are controlled via this arbiter.
    -- 'ON' condition: the machine starts if there is a bus request from this
    -- arbiter, the bus grant input is asserted by the CPU (BGIn = '0'), the 
    -- bus request came from the GLUE (BR_In = '0'), the bus is free(ASn and
    -- DTACKn are not asserted) and there is no bus access by other devices
    -- (BGACKn = '1').
    -- 'OFF' condition: the machine stops, if the RDYn is asserted by the DMA
    -- chip during the ASn is controlled high by this arbiter.
	begin
		if RESn = '0' then
			DMA_PHASE <= IDLE;
		elsif CLK = '1' and CLK' event then
			case DMA_PHASE is
				when IDLE => -- Wait for arbitrated bus.
                    if BR_In = '0' and BGIn = '0' and BGACK_INn = '1' and AS_INn = '1' and DTACKn = '1' then
                        DMA_PHASE <= SYNC;
                    else
                        DMA_PHASE <= IDLE;
                    end if;
                when SYNC => -- Synchronize to MCU.
                    if DMA_MODE = '1' and SYNC_CNT_N = "00" then -- Write to target, read from RAM.
                        DMA_PHASE <= READ;
                    elsif SYNC_CNT_N = "00" then -- Read from target, write to RAM.
                        DMA_PHASE <= WRITE;
                    else
                        DMA_PHASE <= SYNC;
                    end if;
				when READ =>
                    if RDY_INn = '0' then -- Break immediately, due to tight timing.
						DMA_PHASE <= IDLE;
					else
						DMA_PHASE <= READ;
					end if;
				when WRITE =>
                    if RDY_INn = '0' then -- Break immediately, due to tight timing.
						DMA_PHASE <= IDLE;
					else
						DMA_PHASE <= WRITE;
					end if;
			end case;
		end if;
	end process DMA_CTRL;

    MCU_SYNC_N: process
    -- This counter is responsible to enter the DMA sequence in the
    -- correct MCU phase. This counter detects the phase depending on
    -- the DTACKn signal during access to the RAM. If there is a DMA
    -- sequence, the DMA_PHASE READ or WRITE is activated depending
    -- on this counter. Once active, the timing is depending on the
    -- two counters SLICE_CNT in this arbiter logic and TIME_SLICE_CNT
    -- in the MCU controller logic which are then in a correct phase
    -- correlation during the DMA cycle. This logic requires that the
    -- RAMn signal in the GLUE address decoder is generated without
    -- the ASn signal for DMA access. This principle works fine as
    -- long as the memory sharing is equal 50% for video access and
    -- 50% for CPU and coprocessor access. To achieve the correct
    -- timing, this process is using the negative clock edge.
	variable LOCK   : boolean;
	begin
        wait until CLK = '0' and CLK' event;
        if DMA_PHASE = IDLE and RAMn = '0' and DTACKn = '0' and LOCK = false then
            SYNC_CNT_N <= "00";
            LOCK := true;
        elsif DTACKn = '1' then
            LOCK := false;
            SYNC_CNT_N <= SYNC_CNT_N + '1';
        else
            SYNC_CNT_N <= SYNC_CNT_N + '1';
		end if;
	end process MCU_SYNC_N;

	P_TIME_SLICES: process(RESn, CLK)
    -- This counter provides the correct timing information
    -- to establish a synchronous bus operation with the MCU.
    -- This results in the demand of a synchronisation with
    -- the MCU counter, when entering the DMA operation. See
    -- also the explanation to the MCU_SYNC_N in this file.
	begin
		if RESn = '0' then
			SLICE_CNT <= "111"; -- S7.
        elsif CLK = '1' and CLK' event then
            if DMA_PHASE = IDLE or DMA_PHASE = SYNC then
                SLICE_CNT <= "111"; -- DMA inactive.
            elsif DMA_PHASE = READ and RDY_INn = '0' then
                SLICE_CNT <= "111"; -- RDY_INn comes in S6, S7.
            elsif DMA_PHASE = WRITE and RDY_INn = '0' then
                SLICE_CNT <= "111"; -- RDY_INn comes in S0, S1.
            elsif SLICE_CNT = "011" then
                SLICE_CNT <= "000"; -- DMA active, re-enter next cycle.
            else
                SLICE_CNT <= SLICE_CNT + '1';
            end if;
		end if;
	end process P_TIME_SLICES;

	T_SLICE <=  S0 when SLICE_CNT = "000" and CLK = '1' else
			    S1 when SLICE_CNT = "000" and CLK = '0' else
			    S2 when SLICE_CNT = "001" and CLK = '1' else
			    S3 when SLICE_CNT = "001" and CLK = '0' else
			    S4 when SLICE_CNT = "010" and CLK = '1' else
			    S5 when SLICE_CNT = "010" and CLK = '0' else
			    S6 when SLICE_CNT = "011" and CLK = '1' else
			    S7 when SLICE_CNT = "011" and CLK = '0' else IDLE;

    -- This is the timing information equal to the CPU or Blitter bus timing.
	with T_SLICE select AS_OUTn   <= '0'    when           S2 | S3 | S4 | S5 | S6, '1' when others;
 	with T_SLICE select UDS_RD_EN <= '1'    when           S2 | S3 | S4 | S5 | S6, '0' when others;
	with T_SLICE select LDS_RD_EN <= '1'    when           S2 | S3 | S4 | S5 | S6, '0' when others;    
	with T_SLICE select UDS_WR_EN <= '1'    when                     S4 | S5 | S6, '0' when others;
	with T_SLICE select LDS_WR_EN <= '1'    when                     S4 | S5 | S6, '0' when others;
	with T_SLICE select RDY_OUTn  <= '0'    when S0 | S1 | S2 | S3               , '1' when others;

	UDS_OUTn <= '0' when UDS_RD_EN = '1' and DMA_PHASE = READ else
                '0' when UDS_WR_EN = '1' and DMA_PHASE = WRITE else '1';
    
	LDS_OUTn <= '0' when LDS_RD_EN = '1' and DMA_PHASE = READ else
                '0' when LDS_WR_EN = '1' and DMA_PHASE = WRITE else '1';

	RWn_OUT <= '0' when DMA_PHASE = WRITE else '1';

    FC_OUT <= "001"; -- Declare the DMA data as user data.
	
    -- This enables the bus control signals in DMA mode.
    CTRL_EN <= '1' when DMA_PHASE /= IDLE and DMA_PHASE /= SYNC else '0';
end BEHAVIOR;
