------------------------------------------------------------------------
----                                                                ----
---- ATARI Falcon VIDEL compatible IP Core                          ----
----                                                                ----
---- This file is part of the SUSKA ATARI clone project.            ----
---- http://www.experiment-s.de                                     ----
----                                                                ----
---- Description:                                                   ----
----   Atari VIDEL compatible IP core. Refer to the Atari Falcon    ----
----   documentation for further information. There are important   ----
----   informations in the following documents:                     ----
----   1. Atari-Falcon030-Service-Guide.                            ----
----   2. The Authoritative Guide To The Falcon Video Hardware.     ----
----   3. Falcon030_Tech_Doc_10-1-1992.                             ----
----   4. Falcon030DeveloperDocumentation.                          ----
----   5. Atari-Falcon030-Developer-Support-Package.                ----
----                                                                ----
---- This is the video core providing video data management and     ----
---- video timing.                                                  ----
----                                                                ----
---- Remarks:                                                       ----
---- The VIDEO_CTRL bit 7, means the video data bus width is not    ----
---- supported in this core. A 32 bit video data bus is assumed.    ----
----                                                                ----
---- Author(s):                                                     ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de     ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Copyright  2014 Wolfgang Foerster Inventronik GmbH.           ----
---- All rights reserved. No portion of this sourcecode may be      ----
---- reproduced or transmitted in any form by any means, whether    ----
---- by electronic, mechanical, photocopying, recording or          ----
---- otherwise, without my written permission.                      ----
----                                                                ----
------------------------------------------------------------------------

-- Revision History
-- Revision 2K14B  20141224 WF
--   Initial Release.
-- Revision 2K15B  20141224 WF
--   Removed the F_PALETTE instance.
--   Replaced conv_integer by to_integer.
--

library work;
use work.VIDEL_PKG.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

entity VIDEO_CORE is
    port(
        CLK_32M0        : in std_logic; -- 32MHz clock input.
        CLK_25M175      : in std_logic; -- 52.175MHz clock input.
        CLK_EXT         : in std_logic; -- External clock input.
        RESET           : in std_logic;
        RWn             : in std_logic;
        VCS             : in std_logic; -- Videl chip select.
        ADR             : in std_logic_vector(11 downto 1);
        DATA_IN         : in std_logic_vector(15 downto 0);
        DATA_OUT        : out std_logic_vector(15 downto 0);
        DATA_EN         : out std_logic;

        VDATA_IN        : in std_logic_vector(31 downto 0);
        DE_OUT          : out std_logic;
        VDATA_REQ       : out std_logic;
        VDATA_ACK       : in std_logic;
        EVENn_ODD       : out std_logic;

        HSYNC           : out std_logic; -- Horizontal sync pulse.
        HSYNC_EN        : out std_logic;
        HSYNC_POL       : out std_logic;
        VSYNC           : out std_logic; -- Vertical sync pulse.
        VSYNC_EN        : out std_logic;
        VSYNC_POL       : out std_logic;
        CSYNC           : out std_logic; -- Composite video sync signal.
        COLOR           : out std_logic; -- Composite video sync signal for NTSC.

        HINT            : out std_logic; -- Vertical interrupt strobe.
        VINT            : out std_logic; -- Vertical interrupt strobe.

        DOTCK           : out std_logic;

        MONO            : out std_logic;
        
        R               : out std_logic_vector(7 downto 0);
        G               : out std_logic_vector(7 downto 0);
        B               : out std_logic_vector(7 downto 0)      
    );
end entity VIDEO_CORE;

architecture BEHAVIOR of VIDEO_CORE is
type VTYPES is(FALCON, ST);
type F_PALETTE_TYPE is array(0 to 255) of std_logic_vector(15 downto 0);
type ST_PALETTE_TYPE is array(0 to 15) of std_logic_vector(11 downto 0);
type SHIFTREG_TYPE is array(0 to 7) of std_logic_vector(15 downto 0);
type VBUFFER_TYPE is array(0 to 16) of std_logic_vector(31 downto 0);
signal F_PALETTE_HI         : F_PALETTE_TYPE;
signal F_PALETTE_LO         : F_PALETTE_TYPE;
signal ST_PALETTE           : ST_PALETTE_TYPE;
signal VTYPE                : VTYPES;
signal ADR_I                : std_logic_vector(11 downto 0);
signal SYNCMODE             : std_logic;
signal F_PALETTE_WREN_HI    : std_logic;
signal F_PALETTE_WREN_LO    : std_logic;
signal F_PALETTE_ADR        : std_logic_vector(7 downto 0);
signal F_PALETTE_DOUT_HI    : std_logic_vector(15 downto 0);
signal F_PALETTE_DOUT_LO    : std_logic_vector(15 downto 0);
signal SHIFTMODE_ST         : std_logic_vector(1 downto 0);
signal SHIFTMODE_F          : std_logic_vector(10 downto 0);
signal HSCROLL              : std_logic_vector(3 downto 0);
signal VIDEO_CTRL           : std_logic_vector(8 downto 0);
signal VMODE                : std_logic_vector(3 downto 0);
signal HHC                  : std_logic_vector(8 downto 0); -- Horizontal Hold Counter.
signal HHT                  : std_logic_vector(8 downto 0); -- Horizontal Hold Timer.
signal HBB                  : std_logic_vector(8 downto 0); -- Horizontal Boarder Begin.
signal HBE                  : std_logic_vector(8 downto 0); -- Horizontal Boarder End.
signal HDB                  : std_logic_vector(9 downto 0); -- Horizontal Display begin.
signal HDE                  : std_logic_vector(8 downto 0); -- Horizontal Display End.
signal HSS                  : std_logic_vector(8 downto 0); -- Horizontal Sync Start.
signal HFS                  : std_logic_vector(8 downto 0); -- Horizontal Frame Start.
signal HEE                  : std_logic_vector(8 downto 0); -- Horizontal EE.
signal VFC                  : std_logic_vector(10 downto 0); -- Vertical Frequency Counter.
signal VFT                  : std_logic_vector(10 downto 0); -- Vertical Frequency Timer.
signal VBB                  : std_logic_vector(10 downto 0); -- Vertical Boarder begin.
signal VBE                  : std_logic_vector(10 downto 0); -- Vertical Boarder End.
signal VDB                  : std_logic_vector(10 downto 0); -- Vertical Display Begin.
signal VDE                  : std_logic_vector(10 downto 0); -- Vertical DisplAy End.
signal VSS                  : std_logic_vector(10 downto 0); -- Vertical Sync Start.
signal SEL16                : integer range 0 to 15;
signal R_PALETTE            : std_logic_vector(7 downto 0);
signal G_PALETTE            : std_logic_vector(7 downto 0);
signal B_PALETTE            : std_logic_vector(7 downto 0);
signal VIDEO_MODE           : VIDEO_MODES;
signal VIDEO_BUFFER         : VBUFFER_TYPE;
signal PALETTE_SEL          : std_logic_vector(7 downto 0); -- Selector for palette registers.
signal MONO_F               : std_logic;
signal SHFT_REQ             : std_logic;
signal VBASE_CLK            : std_logic;
signal VIDEO_STRB           : std_logic;
signal SHFT_STRB            : std_logic;
signal HILOn                : std_logic; -- HILOn = '0' is the first half line.
signal BOARDER              : std_logic;
signal DE                   : std_logic;
signal HBASE_OFFS           : std_logic_vector(7 downto 0);
signal VF_50_60n            : std_logic;
signal SR_00                : std_logic;
signal HSYNC_I              : std_logic;
signal VSYNC_I              : std_logic;
begin
    DOTCK <= VBASE_CLK;
    VBASE_CLK <= CLK_EXT when SYNCMODE = '1' else
                 CLK_32M0 when VIDEO_CTRL(2) = '0' else CLK_25M175; -- Video base clock, '0' = 32MHz - '1' = 25.175MHz.

    ADR_I <= ADR & '0';

    F_PALETTE_WREN_HI <= '1' when VCS = '1' and RWn = '0' and ADR_I >= x"801" and ADR_I(3 downto 2) < x"C00" and ADR_I(1) = '0' else '0';
    F_PALETTE_WREN_LO <= '1' when VCS = '1' and RWn = '0' and ADR_I >= x"801" and ADR_I(3 downto 2) < x"C00" and ADR_I(1) = '1' else '0';
    F_PALETTE_ADR <= ADR(9 downto 2) when VCS = '1' else -- Bus access.
                     PALETTE_SEL when VIDEO_MODE = F_8BITPL else 
                     SHIFTMODE_F(3 downto 0) & PALETTE_SEL(3 downto 0); -- Colour bank selection & palette selection.

    PALETTE_F: process(CLK_32M0, F_PALETTE_HI, F_PALETTE_LO)
    -- This process is written in that manner, that 8192 bits
    -- RAM will be inferred.
    variable F_PALETTE_ADR_PNTR   : integer range 0 to 255;
    begin
        if CLK_32M0 = '1' and CLK_32M0' event then
            F_PALETTE_ADR_PNTR := To_Integer(unsigned(F_PALETTE_ADR));
            if F_PALETTE_WREN_HI = '1' then
                F_PALETTE_HI(F_PALETTE_ADR_PNTR) <= DATA_IN;
            end if;
            --
            if F_PALETTE_WREN_LO = '1' then
                F_PALETTE_LO(F_PALETTE_ADR_PNTR) <= DATA_IN;
            end if;        
        end if;
        F_PALETTE_DOUT_HI <= F_PALETTE_HI(F_PALETTE_ADR_PNTR);
        F_PALETTE_DOUT_LO <= F_PALETTE_LO(F_PALETTE_ADR_PNTR);
    end process PALETTE_F;
    
    REGS: process
    begin
        wait until CLK_32M0 = '1' and CLK_32M0' event;
        if RESET = '1' then
            SYNCMODE <= '0';
            SHIFTMODE_ST <= (others => '0');
            SHIFTMODE_F <= (others => '0');
            HSCROLL <= (others => '0');
            VIDEO_CTRL <= (others => '0');
            VMODE <= (others => '0');
            ST_PALETTE <= (others => (others => '0'));
            HHT <= (others => '0');
            HBB <= (others => '0');
            HBE <= (others => '0');
            HDB <= (others => '0');
            HDE <= (others => '0');
            HSS <= (others => '0');
            HFS <= (others => '0');
            HEE <= (others => '0');
            VFT <= (others => '0');
            VBB <= (others => '0');
            VBE <= (others => '0');
            VDB <= (others => '0');
            VDE <= (others => '0');
            VSS <= (others => '0');
        end if;
        --
        if VCS = '1' and ADR_I = x"20A" and RWn = '0' then -- x"FFFF820A".
            SYNCMODE <= DATA_IN(8);
        end if;
        --
        if VCS = '1' and RWn = '0' then
            case ADR_I is
                when x"25E" => ST_PALETTE(15) <= DATA_IN(11 downto 0); -- x"FFF825E - FFFF825F"
                when x"25C" => ST_PALETTE(14) <= DATA_IN(11 downto 0);
                when x"25A" => ST_PALETTE(13) <= DATA_IN(11 downto 0);
                when x"258" => ST_PALETTE(12) <= DATA_IN(11 downto 0);
                when x"256" => ST_PALETTE(11) <= DATA_IN(11 downto 0);
                when x"254" => ST_PALETTE(10) <= DATA_IN(11 downto 0);
                when x"252" => ST_PALETTE(9) <= DATA_IN(11 downto 0);
                when x"250" => ST_PALETTE(8) <= DATA_IN(11 downto 0);
                when x"24E" => ST_PALETTE(7) <= DATA_IN(11 downto 0);
                when x"24C" => ST_PALETTE(6) <= DATA_IN(11 downto 0);
                when x"24A" => ST_PALETTE(5) <= DATA_IN(11 downto 0);
                when x"248" => ST_PALETTE(4) <= DATA_IN(11 downto 0);
                when x"246" => ST_PALETTE(3) <= DATA_IN(11 downto 0);
                when x"244" => ST_PALETTE(2) <= DATA_IN(11 downto 0);
                when x"242" => ST_PALETTE(1) <= DATA_IN(11 downto 0);
                when x"240" => ST_PALETTE(0) <= DATA_IN(11 downto 0); -- x"FFF8240 - FFFF8241"
                when others =>
                    null;
            end case;
        end if;
        --
        if VCS = '1' and ADR_I = x"260" and RWn = '0' then -- x"FFFF8260"
            SHIFTMODE_ST <= DATA_IN(9 downto 8);
        end if;
        --
        if VCS = '1' and ADR_I = x"264" and RWn = '0' then -- x"FFFF8265"
            HSCROLL <= DATA_IN(3 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"266" and RWn = '0' then -- x"FFFF8266 - FFFF8267"
            SHIFTMODE_F <= DATA_IN(10 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"282" and RWn = '0' then -- x"FFFF8282 - FFFF8283"
            HHT <= DATA_IN(8 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"284" and RWn = '0' then -- x"FFFF8284 - FFFF8285"
            HBB <= DATA_IN(8 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"286" and RWn = '0' then -- x"FFFF8286 - FFFF8287"
            HBE <= DATA_IN(8 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"288" and RWn = '0' then -- x"FFFF8288 - FFFF8289"
            HDB <= DATA_IN(9 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"28A" and RWn = '0' then -- x"FFFF828A - FFFF828B"
            HDE <= DATA_IN(8 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"28C" and RWn = '0' then -- x"FFFF828C - FFFF828D"
            HSS <= DATA_IN(8 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"28E" and RWn = '0' then -- x"FFFF828E - FFFF828F"
            HFS <= DATA_IN(8 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"290" and RWn = '0' then -- x"FFFF8290 - FFFF82891"
            HEE <= DATA_IN(8 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"2A2" and RWn = '0' then -- x"FFFF82A2 - FFFF82A3"
            VFT <= DATA_IN(10 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"2A4" and RWn = '0' then -- x"FFFF82A4 - FFFF82A5"
            VBB <= DATA_IN(10 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"2A6" and RWn = '0' then -- x"FFFF82A6 - FFFF82A7"
            VBE <= DATA_IN(10 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"2A8" and RWn = '0' then -- x"FFFF82A8 - FFFF82A9"
            VDB <= DATA_IN(10 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"2AA" and RWn = '0' then -- x"FFFF82AA - FFFF82AB"
            VDE <= DATA_IN(10 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"2AC" and RWn = '0' then -- x"FFFF82AC - FFFF82AD"
            VSS <= DATA_IN(10 downto 0);
        end if;
        --
        if VCS = '1' and ADR_I = x"2C0" and RWn = '0' then -- x"FFFF82C0 - FFFF82C1"
            VIDEO_CTRL(8 downto 2) <= DATA_IN(8 downto 2);
        end if;
        --
        if VCS = '1' and ADR_I = x"006" and RWn = '0' then -- x"FFFF8006.
            VIDEO_CTRL(1 downto 0) <= DATA_IN(15 downto 14);
        end if;
        --
        if VCS = '1' and ADR_I = x"2C2" and RWn = '0' then -- x"FFFF82C2 - FFFF82C3"
            VMODE <= DATA_IN(3 downto 0);
        elsif VCS = '1' and ADR_I = x"260" and RWn = '0' then -- Writing the ST shift mode register alters the VMODE.
            case DATA_IN(9 downto 8) is
                when "00" => VMODE <= x"0"; -- ST low.
                when "01" => VMODE <= x"4"; -- ST medium.
                when "10" => VMODE <= x"6"; -- ST low.
                when "11" => null; -- Reserved.
            end case;
        end if;
        --
    end process REGS;

    -- Unused bits read back as '0';
    DATA_OUT <= x"00" & "000000" & VF_50_60n & SYNCMODE when VCS = '1' and ADR_I = x"20A" and RWn = '1' else 
                x"0" & ST_PALETTE(0) when VCS = '1' and ADR_I = x"240" and RWn = '1' else
                x"0" & ST_PALETTE(1) when VCS = '1' and ADR_I = x"242" and RWn = '1' else
                x"0" & ST_PALETTE(2) when VCS = '1' and ADR_I = x"244" and RWn = '1' else
                x"0" & ST_PALETTE(3) when VCS = '1' and ADR_I = x"246" and RWn = '1' else
                x"0" & ST_PALETTE(4) when VCS = '1' and ADR_I = x"248" and RWn = '1' else
                x"0" & ST_PALETTE(5) when VCS = '1' and ADR_I = x"24A" and RWn = '1' else
                x"0" & ST_PALETTE(6) when VCS = '1' and ADR_I = x"24C" and RWn = '1' else
                x"0" & ST_PALETTE(7) when VCS = '1' and ADR_I = x"24E" and RWn = '1' else
                x"0" & ST_PALETTE(8) when VCS = '1' and ADR_I = x"250" and RWn = '1' else
                x"0" & ST_PALETTE(9) when VCS = '1' and ADR_I = x"252" and RWn = '1' else
                x"0" & ST_PALETTE(10) when VCS = '1' and ADR_I = x"254" and RWn = '1' else
                x"0" & ST_PALETTE(11) when VCS = '1' and ADR_I = x"256" and RWn = '1' else
                x"0" & ST_PALETTE(12) when VCS = '1' and ADR_I = x"258" and RWn = '1' else
                x"0" & ST_PALETTE(13) when VCS = '1' and ADR_I = x"25A" and RWn = '1' else
                x"0" & ST_PALETTE(14) when VCS = '1' and ADR_I = x"25C" and RWn = '1' else
                x"0" & ST_PALETTE(15) when VCS = '1' and ADR_I = x"25E" and RWn = '1' else
                "000000" & SHIFTMODE_ST & x"00" when VCS = '1' and ADR_I = x"260" and RWn = '1' else
                x"0" & HSCROLL & x"0" & HSCROLL when VCS = '1' and ADR_I = x"264" and RWn = '1' else -- Left is Shadow.
                "00000" & SHIFTMODE_F when VCS = '1' and ADR_I = x"266" and RWn = '1' else
                "0000000" & HHC when VCS = '1' and ADR_I = x"280" and RWn = '1' else
                "0000000" & HHT when VCS = '1' and ADR_I = x"282" and RWn = '1' else
                "0000000" & HBB when VCS = '1' and ADR_I = x"284" and RWn = '1' else
                "0000000" & HBE when VCS = '1' and ADR_I = x"286" and RWn = '1' else
                "000000" & HDB when VCS = '1' and ADR_I = x"288" and RWn = '1' else
                "0000000" & HDE when VCS = '1' and ADR_I = x"28A" and RWn = '1' else
                "0000000" & HSS when VCS = '1' and ADR_I = x"28C" and RWn = '1' else
                "0000000" & HFS when VCS = '1' and ADR_I = x"28E" and RWn = '1' else
                "0000000" & HEE when VCS = '1' and ADR_I = x"290" and RWn = '1' else
                "00000" & VFC when VCS = '1' and ADR_I = x"2A0" and RWn = '1' else
                "00000" & VFT when VCS = '1' and ADR_I = x"2A2" and RWn = '1' else
                "00000" & VBB when VCS = '1' and ADR_I = x"2A4" and RWn = '1' else
                "00000" & VBE when VCS = '1' and ADR_I = x"2A6" and RWn = '1' else
                "00000" & VDB when VCS = '1' and ADR_I = x"2A8" and RWn = '1' else
                "00000" & VDE when VCS = '1' and ADR_I = x"2AA" and RWn = '1' else
                "00000" & VSS when VCS = '1' and ADR_I = x"2AC" and RWn = '1' else
                "0000000" & VIDEO_CTRL when VCS = '1' and ADR_I = x"2C0" and RWn = '1' else
                x"000" & VMODE when VCS = '1' and ADR_I = x"2C2" and RWn = '1' else
                F_PALETTE_DOUT_HI when VCS = '1' and ADR_I >= x"801" and ADR_I < x"C00" and ADR_I(1) = '0' and RWn = '1' else F_PALETTE_DOUT_LO;

    DATA_EN <=  '1' when VCS = '1' and RWn = '1' else '0';

    VTYPE_SWITCH: process
    -- This Flip Flop determines whether the video system
    -- works in Falcon or in ST Mode.
    begin
        wait until CLK_32M0 = '1' and CLK_32M0' event;
        if RESET = '1' then
            VTYPE <= ST;
        elsif VCS = '1' and ADR_I = x"260" and RWn = '0' then -- SHIFTMODE_ST write access.
            if SHIFTMODE_F(10) = '0' and SHIFTMODE_F(8) = '0' and SHIFTMODE_F(4) = '0' then
                VTYPE <= ST;
            end if;
        elsif VCS = '1' and ADR_I = x"266" and RWn = '0' then -- SHIFTMODE_F write access.
            VTYPE <= FALCON;
        end if;
    end process VTYPE_SWITCH;

    SEL16 <= Conv_Integer(PALETTE_SEL(3 downto 0));
        
    -- Control outputs:
    R_PALETTE <= ST_PALETTE(SEL16)(11 downto 8) & x"0" when VTYPE = ST else F_PALETTE_DOUT_HI(15 downto 8);
    G_PALETTE <= ST_PALETTE(SEL16)(7 downto 4) & x"0" when VTYPE = ST else F_PALETTE_DOUT_HI(7 downto 0);
    B_PALETTE <= ST_PALETTE(SEL16)(3 downto 0) & x"0" when VTYPE = ST else F_PALETTE_DOUT_LO(7 downto 0);

    VIDEO_MODE <= F_MONO when SHIFTMODE_F(10) = '1'  and VTYPE = FALCON else
                  F_TRUEC when SHIFTMODE_F(8) = '1'  and VTYPE = FALCON else
                  F_8BITPL when SHIFTMODE_F(4) = '1'  and VTYPE = FALCON else
                  F_4BITPL when VTYPE = FALCON else
                  STE_MID when SHIFTMODE_ST = "01" else
                  STE_LOW when SHIFTMODE_ST = "10" else STE_MONO;

    VF_50_60n <= '0' when VIDEO_CTRL(1 downto 0) = "00" else '1';

    VFIFO: process
    -- This is the video data FIFO buffer. it works as follows:
    -- The FIFO is loaded from the VDATA bus each time a VDATA_ACK indicates new video data. The bus
    -- width of VDATA is 32 bit. The loading increments a load pointer. Each Video data request forces
    -- 17 32bit wide long words to be loadeds into this FIFO (in a burst) unaffected of the readout by
    -- video data requests of the shifter or the true colour mode. The VDATA_REQ is controlled in a way,
    -- that an overflow of FIFO should never appear. Nevertheless, the FIFO logic permits a further
    -- loading of video data, if it is full. This condition is indicated by the L32_PNTR value of 17.
    -- In case of a video shifter or true colour video data request, the VDATA_ACK signal is delayed by
    -- one clock cycle and the loading of the VDATA_IN takes place right after loading the shift register
    -- or true colour VIDEO_BUFFER.
    -- There are two independant pointers, the L32_PNTR which indicates the current free FIFO position
    -- and the LOAD_PNTR which indicates the value of the currently loaded 32 Bit video data during a
    -- burst access.
    variable LOAD_PNTR      : integer range 0 to 17;
    variable L32_PNTR       : integer range 0 to 17;
    variable VDATA_ACK_D    : std_logic;
    variable WORDSWAP       : boolean;
    begin
-- L32_PNTR bis 16?
        wait until VBASE_CLK = '1' and VBASE_CLK' event;
        if (VIDEO_MODE = F_TRUEC and DE = '1' and SHFT_STRB = '1') or (VIDEO_MODE /= F_TRUEC and SHFT_REQ = '1') then
            VDATA_ACK_D := VDATA_ACK;
            --
            case VIDEO_MODE is
                when F_8BITPL => -- Eight bitplanes.
                    if L32_PNTR > 3 then
                        for i in 0 to 12 loop
                            VIDEO_BUFFER(i) <= VIDEO_BUFFER(i + 4);
                        end loop;
                        L32_PNTR := L32_PNTR - 4;
                    end if;
                when F_4BITPL | STE_LOW => -- Four Bitplanes.
                    if L32_PNTR > 1 then
                        for i in 0 to 14 loop
                            VIDEO_BUFFER(i) <= VIDEO_BUFFER(i + 2);
                        end loop;
                        L32_PNTR := L32_PNTR - 2;
                    end if;
                when STE_MID => -- Two Bitplanes.
                    if L32_PNTR > 0 then
                        for i in 0 to 15 loop
                            VIDEO_BUFFER(i) <= VIDEO_BUFFER(i + 1);
                        end loop;
                        L32_PNTR := L32_PNTR - 1;
                    end if;
                when F_MONO | STE_MONO | F_TRUEC => -- We wrap around one word.
                    if L32_PNTR > 0 then
                        if WORDSWAP = false then                        
                            VIDEO_BUFFER(0) <= VIDEO_BUFFER(0)(15 downto 0) & x"0000";
                            WORDSWAP := true;
                        else
                            for i in 0 to 15 loop
                                VIDEO_BUFFER(i) <= VIDEO_BUFFER(i + 1);
                            end loop;
                            L32_PNTR := L32_PNTR - 1;
                            WORDSWAP := false;
                        end if;                        
                    end if;
            end case;
        elsif (VDATA_ACK = '1' or VDATA_ACK_D = '1') and L32_PNTR < 17 then
            for i in 0 to 16 loop
                if i = L32_PNTR then
                    VIDEO_BUFFER(i) <= VDATA_IN;
                end if;
            end loop;
            L32_PNTR := L32_PNTR + 1;
            VDATA_ACK_D := '0';
        end if;
        --
        if (VDATA_ACK = '1' or VDATA_ACK_D = '1') and L32_PNTR < 33 then
            LOAD_PNTR := LOAD_PNTR + 1;
        elsif L32_PNTR = 0 then
            LOAD_PNTR := 0;
        end if;
        
        if L32_PNTR = 0 then
            VDATA_REQ <= '1';
        elsif LOAD_PNTR = 17 then
            VDATA_REQ <= '0';
        end if;
    end process VFIFO;

    P_SHIFTER: process
    variable SHIFTREGS  : SHIFTREG_TYPE;
    variable PIX_CNT    : integer range 0 to 15;
    variable H_SHIFT    : integer range 0 to 15;
    begin
        wait until VBASE_CLK = '1' and VBASE_CLK' event;
        if DE = '1' and SHFT_STRB = '1' and PIX_CNT = 0 then
            SHIFTREGS(7) := VIDEO_BUFFER(3)(15 downto 0);
            SHIFTREGS(6) := VIDEO_BUFFER(3)(31 downto 16);
            SHIFTREGS(5) := VIDEO_BUFFER(2)(15 downto 0);
            SHIFTREGS(4) := VIDEO_BUFFER(2)(31 downto 16);
            SHIFTREGS(3) := VIDEO_BUFFER(1)(15 downto 0);
            SHIFTREGS(2) := VIDEO_BUFFER(1)(31 downto 16);
            SHIFTREGS(1) := VIDEO_BUFFER(0)(15 downto 0);
            SHIFTREGS(0) := VIDEO_BUFFER(0)(31 downto 16);
            --
            PIX_CNT := 15;
        elsif DE = '1' and SHFT_STRB = '1' then
            for i in 7 downto 0 loop
                SHIFTREGS(i) := SHIFTREGS(i)(14 downto 0) & '0'; -- Shift right.
            end loop;
            PIX_CNT := PIX_CNT - 1;
        end if;
        --
        case PIX_CNT is
            when 0 => SHFT_REQ <= '1';
            when others => SHFT_REQ <= '0';
        end case;
        --
        H_SHIFT := To_Integer(unsigned(HSCROLL));
        for i in 7 downto 0 loop
            PALETTE_SEL(i) <= SHIFTREGS(i)(H_SHIFT);
        end loop;
        --
        MONO_F <= SHIFTREGS(0)(0) xor F_PALETTE_DOUT_LO(0);
        SR_00 <= SHIFTREGS(0)(0);
    end process P_SHIFTER;

    -- Monochrome video: this signal is used for video information in the
    -- monochrome modes. It is used as overlay switch in the true colour
    -- video mode. If overlay is active in the true colour mode, bit 5 
    -- of the video buffer determines the insertion of external video.
    -- See the Developer Support Package for the Atari Falcon030 for
    -- further information.
    MONO <= MONO_F when VIDEO_MODE <= F_MONO else
            SR_00 xor ST_PALETTE(0)(0) when VIDEO_MODE <= STE_MONO else 
            '1' when VIDEO_MODE = F_TRUEC and SHIFTMODE_F(9) = '1' and VIDEO_BUFFER(0)(5) = '1' else '0'; -- Overlay.

    -- True colour is RrrrrGgggggBbbbb.
    R <= x"00" when BOARDER = '1' else
         VIDEO_BUFFER(0)(15 downto 11) & "000" when VIDEO_MODE = F_TRUEC else R_PALETTE;
    G <= x"00" when BOARDER = '1' else
         VIDEO_BUFFER(0)(10 downto 5) & "00" when VIDEO_MODE = F_TRUEC else G_PALETTE;
    B <= x"00" when BOARDER = '1' else 
         VIDEO_BUFFER(0)(4 downto 0) & "000" when VIDEO_MODE = F_TRUEC else B_PALETTE;

    VIDEO_STROBES: process (VBASE_CLK, VMODE, VIDEO_CTRL)
    variable VCNT   : std_logic_vector(1 downto 0);
    begin
        if VBASE_CLK = '1' and VBASE_CLK' event then
            VCNT := VCNT + '1';
        end if;
        --
        if VIDEO_CTRL(1 downto 0) = "10" then -- VGA monitor.
            case VMODE(3 downto 2) is
                when "00" =>
                    case VCNT is
                        when "00" => VIDEO_STRB <= '1'; -- VBASE_CLK/4.
                        when others => VIDEO_STRB <= '0';
                    end case;
                when "01" | "10" =>
                    case VCNT is
                        when "00" | "10" => VIDEO_STRB <= '1'; -- VBASE_CLK/2.
                        when others => VIDEO_STRB <= '0';
                    end case;
                when "11" => VIDEO_STRB <= '0'; -- Off.
            end case;
        else -- All other video modes.
            case VMODE(3 downto 2) is
                when "00" =>
                    case VCNT is
                        when "00" => VIDEO_STRB <= '1'; -- VBASE_CLK/4.
                        when others => VIDEO_STRB <= '0';
                    end case;
                when "01" =>
                    case VCNT is
                        when "00" | "10" => VIDEO_STRB <= '1'; -- VBASE_CLK/2.
                        when others => VIDEO_STRB <= '0';
                    end case;
                when "10" => VIDEO_STRB <= '1';
                when "11" => VIDEO_STRB <= '0'; -- Off.
            end case;
        end if;
        --
        case VMODE(3 downto 2) is
            when "00" =>
                case VCNT is
                    when "00" => SHFT_STRB <= '1'; -- 4 cycles.
                    when others => SHFT_STRB <= '0';
                end case;
            when "01" =>
                case VCNT is
                    when "00" | "10" => SHFT_STRB <= '1'; -- 2 cycles.
                    when others => SHFT_STRB <= '0';
                end case;
            when "10" => SHFT_STRB <= '1'; -- One cycle.
            when "11" => SHFT_STRB <= '0'; -- Off.
        end case;
    end process VIDEO_STROBES;
    
    HORIZONTAL_TIMING: process
    begin
        wait until VBASE_CLK = '1' and VBASE_CLK' event;
        if VIDEO_STRB = '1' and HHC <= HHT then
            HHC <= HHC + '1';
        elsif VIDEO_STRB = '1' then
            HHC <= (others => '0');
            HILOn <= not HILOn;
        end if;
    end process HORIZONTAL_TIMING;

    VERTICAL_TIMING: process
    variable EnO_VAR    : std_logic;
    begin
        wait until VBASE_CLK = '1' and VBASE_CLK' event;
        if VIDEO_STRB = '1' and HHC = HHT and VFC <= VFT then
            VFC <= VFC + '1';
        elsif VIDEO_STRB = '1' and HILOn = '1' and HHC = HHT then
            VFC <= (others => '0');
            EnO_VAR := not EnO_VAR; 
            EVENn_ODD <= EnO_VAR;
        end if;
    end process VERTICAL_TIMING;

    HSYNC_EN <= SHIFTMODE_F(6);
    HSYNC_POL <= VIDEO_CTRL(6);
    HSYNC <= HSYNC_I;
    HSYNC_I <= '1' when HILOn = '1' and HHC >= HSS else
               '1' when HILOn = '0' and VIDEO_CTRL(3) = '1' and HHC <= HEE and VFC < x"000000110" else -- First five lines.
               '1' when HILOn = '0' and VIDEO_CTRL(3) = '1' and HHC <= HEE and VFT - VFC  < x"000000110" else -- Last five lines.
               '1' when HILOn = '0' and VIDEO_CTRL(3) = '1' and HHC <= HFS else '0'; -- Lines in between.
             
    VSYNC_EN <= SHIFTMODE_F(5);
    VSYNC_POL <= VIDEO_CTRL(5);
    VSYNC <= VSYNC_I;
    VSYNC_I <= '1' when VFC >= VSS else '0';

    BOARDER <= '0' when VIDEO_CTRL(1 downto 0) = "00" else -- Monochrome monitor.
               '1' when VFC <= VBE else
               '1' when VFC >= VBB else
               '1' when HHC <= HBB else
               '1' when HILOn = '1' and HHC >= HBE else '0';

    CSYNC <= not(HSYNC_I or VSYNC_I);
    COLOR <= HSYNC_I or VSYNC_I;

    HBASE_OFFS <= x"80" when VIDEO_CTRL(8)= '0' else x"40"; -- 128 or 64 video cycles.

    DE_OUT <= DE;
    DE <= '0' when VFC <= VDB else
          '0' when VFC >= VDE else
          '0' when HHC <= HDB(8 downto 0) + HBASE_OFFS and HDB(9) = '0'and HILOn = '0' else
          '0' when HHC <= HDB(8 downto 0) + HBASE_OFFS and HDB(9) = '1' and HILOn = '1' else
          '0' when HHC >= HDE + HBASE_OFFS else '1';

    -- VINT is the same as the vertical DE.
    VINT <= '0' when VFC <= VDB else
            '0' when VFC >= VDE else '1';

    -- HINT is the same as the horizontal DE.
    HINT <= '0' when HHC <= HDB(8 downto 0) + HBASE_OFFS and HDB(9) = '0'and HILOn = '0' else
            '0' when HHC <= HDB(8 downto 0) + HBASE_OFFS and HDB(9) = '1' and HILOn = '1' else
            '0' when HHC >= HDE + HBASE_OFFS else '1';          
end architecture BEHAVIOR;
